from abmarl.sim.components.agent import ComponentAgent as Agent
from abmarl.sim.components.state import LifeState


def test_health_agents():
    agents = {
        'agent0': Agent(id='agent0', min_health=0.0, max_health=5.0, initial_health=3.4),
        'agent1': Agent(id='agent1', min_health=0.0, max_health=5.0, initial_health=2.4),
        'agent2': Agent(id='agent2', min_health=0.0, max_health=5.0),
        'agent3': Agent(id='agent3', min_health=0.0, max_health=5.0),
    }

    assert agents['agent0'].min_health == 0.0
    assert agents['agent0'].max_health == 5.0
    assert agents['agent0'].initial_health == 3.4
    assert agents['agent0'].is_alive
    assert agents['agent1'].min_health == 0.0
    assert agents['agent1'].max_health == 5.0
    assert agents['agent1'].initial_health == 2.4
    assert agents['agent1'].is_alive
    assert agents['agent2'].min_health == 0.0
    assert agents['agent2'].max_health == 5.0
    assert agents['agent2'].is_alive
    assert agents['agent3'].min_health == 0.0
    assert agents['agent3'].max_health == 5.0
    assert agents['agent3'].is_alive


def test_life_state():
    agents = {
        'agent0': Agent(id='agent0', min_health=0.0, max_health=5.0, initial_health=3.4),
        'agent1': Agent(id='agent1', min_health=0.0, max_health=5.0, initial_health=2.4),
        'agent2': Agent(id='agent2', min_health=0.0, max_health=5.0),
        'agent3': Agent(id='agent3', min_health=0.0, max_health=5.0),
    }

    assert agents['agent0'].min_health == 0.0
    assert agents['agent0'].max_health == 5.0
    assert agents['agent0'].initial_health == 3.4
    assert agents['agent0'].is_alive
    assert agents['agent1'].min_health == 0.0
    assert agents['agent1'].max_health == 5.0
    assert agents['agent1'].initial_health == 2.4
    assert agents['agent1'].is_alive
    assert agents['agent2'].min_health == 0.0
    assert agents['agent2'].max_health == 5.0
    assert agents['agent2'].is_alive
    assert agents['agent3'].min_health == 0.0
    assert agents['agent3'].max_health == 5.0
    assert agents['agent3'].is_alive

    state = LifeState(agents=agents, entropy=0.5)
    state.reset()

    assert agents['agent0'].health == 3.4
    assert agents['agent1'].health == 2.4
    assert 0.0 <= agents['agent2'].health <= 5.0
    assert 0.0 <= agents['agent3'].health <= 5.0

    state.apply_entropy(agents['agent0'])
    state.apply_entropy(agents['agent1'])
    assert agents['agent0'].health == 2.9
    assert agents['agent1'].health == 1.9

    for _ in range(10):
        state.apply_entropy(agents['agent0'])
        state.apply_entropy(agents['agent1'])
        state.apply_entropy(agents['agent2'])
        state.apply_entropy(agents['agent3'])

    assert not agents['agent0'].is_alive
    assert not agents['agent1'].is_alive
    assert not agents['agent2'].is_alive
    assert not agents['agent3'].is_alive

    state.reset()
    assert agents['agent0'].min_health == 0.0
    assert agents['agent0'].max_health == 5.0
    assert agents['agent0'].initial_health == 3.4
    assert agents['agent0'].is_alive
    assert agents['agent1'].min_health == 0.0
    assert agents['agent1'].max_health == 5.0
    assert agents['agent1'].initial_health == 2.4
    assert agents['agent1'].is_alive
    assert agents['agent2'].min_health == 0.0
    assert agents['agent2'].max_health == 5.0
    assert agents['agent2'].is_alive
    assert agents['agent3'].min_health == 0.0
    assert agents['agent3'].max_health == 5.0
    assert agents['agent3'].is_alive



import numpy as np


class ResourcesDepletedDone:
    """
    Simulation ends when all the resources are depleted.

    resource_state (GridResourceState):
        The state of the resources
    """
    def __init__(self, resource_state=None, **kwargs):
        self.resource_state = resource_state

    def get_done(self, *args, **kwargs):
        """
        Return true if all the resources are depleted.
        """
        return self.get_all_done(**kwargs)

    def get_all_done(self, **kwargs):
        """
        Return True if all the resources are depleted.
        """
        return np.all(self.resource_state.resources == 0)


class DeadDone:
    """
    Dead agents are indicated as done. Additionally, the simulation is over when
    all the agents are dead.

    agents (dict):
        The dictionary of agents.
    """
    def __init__(self, agents=None, **kwargs):
        self.agents = agents

    def get_done(self, agent, **kwargs):
        """
        Return True if the agent is dead. Otherwise, return False.
        """
        return not agent.is_alive

    def get_all_done(self, **kwargs):
        """
        Return True if all agents are dead. Otherwise, return False.
        """
        for agent in self.agents.values():
            if agent.is_alive:
                return False
        return True


class TeamDeadDone:
    """
    Dead agents are indicated as done. Additionally, the simulation is over when
    the only agents remaining are on the same team.

    agents (dict):
        The dictionary of agents.

    number_of_teams (int):
        The fixed number of teams in this simulation.
        Default 0.
    """
    def __init__(self, agents=None, number_of_teams=0, **kwargs):
        self.agents = agents
        assert type(number_of_teams) is int, "number_of_teams must be a positive integer."
        # +1 because team 0 is default team and not counted.
        self.number_of_teams = number_of_teams + 1

    def get_done(self, agent, **kwargs):
        """
        Return True if the agent is dead. Otherwise, return False.
        """
        return not agent.is_alive

    def get_all_done(self, **kwargs):
        """
        Return true if the only agent left alive are all on the same team. Otherwise,
        return false.
        """
        team = np.zeros(self.number_of_teams)
        for agent in self.agents.values():
            if agent.is_alive:
                team[agent.team] += 1
        return sum(team != 0) <= 1


class TooCloseDone:
    """
    Agents that are too close to each other or too close to the edge of the region
    are indicated as done. If any agent is done, the entire simulation is done.

    position (PositionState):
        The position state handler.

    agents (dict):
        Dictionay of agent objects.

    collision_distance (float):
        The threshold for calculating if a collision has occured.

    collision_norm (int):
        The norm to use when calculating the collision. For example, you would
        probably want to use 1 in the Grid space but 2 in a Continuous space.
        Default is 2.
    """
    def __init__(self, position=None, agents=None, collision_distance=None, collision_norm=2,
                 **kwargs):
        assert position is not None
        self.position = position
        self.agents = agents
        assert collision_distance is not None
        self.collision_distance = collision_distance
        self.collision_norm = collision_norm

    def get_done(self, agent, **kwargs):
        """
        Return true if the agent is too close to another agent or too close to
        the edge of the region.
        """
        #  Collision with region edge
        if np.any(agent.position[0] < self.collision_distance) \
                or np.any(agent.position[0] > self.position.region - self.collision_distance) \
                or np.any(agent.position[1] < self.collision_distance) \
                or np.any(agent.position[1] > self.position.region - self.collision_distance):
            return True

        # Collision with other birds
        for other in self.agents.values():
            if other.id == agent.id: continue # Cannot collide with yourself
            if np.linalg.norm(other.position - agent.position, self.collision_norm) < \
                    self.collision_distance:
                return True
        return False

    def get_all_done(self, **kwargs):
        """
        Return true if any agent is too close to another agent or too close to
        the edge of the region.
        """
        for agent in self.agents.values():
            if self.get_done(agent):
                return True
        return False



from abc import ABC, abstractmethod, abstractproperty

from gym.spaces import Box, Dict
import numpy as np

from abmarl.sim.components.agent import HealthObservingAgent, LifeObservingAgent, \
    AgentObservingAgent, PositionObservingAgent, SpeedAngleObservingAgent, \
    VelocityObservingAgent, ResourceObservingAgent, TeamObservingAgent, BroadcastObservingAgent, \
    SpeedAngleAgent, VelocityAgent, BroadcastingAgent, ComponentAgent


class Observer(ABC):
    """
    Base observer class provides the interface required of all observers. Setup
    the agents' observation space according to the Observer's channel.

        agents (dict):
            The dictionary of agents.
    """
    def __init__(self, agents=None, **kwargs):
        self.agents = agents

    def _set_obs_space_simple(self, instance, space_func, **kwargs):
        """
        Observers that don't depend on the type of the other agents can use this
        method.

        instance (Agent):
            An Agent class. This is used in the isinstance check to determine if
            the agent will receive the observation channel.

        space_func (function):
            A function that takes the other agent as input and outputs the
            observation space.
        """
        for agent in self.agents.values():
            if isinstance(agent, instance):
                agent.observation_space[self.channel] = Dict({
                    other.id: space_func(other) for other in self.agents.values()
                })

    def _set_obs_space(self, instance, other_instance, space_func, alt_space_func, **kwargs):
        """
        Observers that depend on the type of the other agents must use this method.

        instance (Agent):
            An Agent class. This is used in the isinstance check to determine if
            the agent will receive the observation channel.

        other_instance (Agent):
            An Agent class. This is used in the isinstance check on the other agents
            to determine whether to use the space_func or the alt_space_func.

        space_func (function):
            A function that takes the other agent as input and outputs the
            observation space.

        alt_space_func (function):
            Use this function for cases when the isinstance check fails on the
            other agent. Function does not have inputs and outputs observation space.
        """
        for agent in self.agents.values():
            if isinstance(agent, instance):
                obs_space = {}
                for other in self.agents.values():
                    if isinstance(other, other_instance):
                        obs_space[other.id] = space_func(other)
                    else:
                        obs_space[other.id] = alt_space_func()
                agent.observation_space[self.channel] = Dict(obs_space)

    def _get_obs(self, agent, instance=None, other_instance=ComponentAgent, attr=None, **kwargs):
        """
        Many observers just directly query the corresponding state field from the
        agent. This function does exactly that, checking the instance of the observing
        agent and the other agents and setting the observation value accordingly.
        """
        if isinstance(agent, instance):
            obs = {}
            for other in self.agents.values():
                if isinstance(other, other_instance):
                    obs[other.id] = getattr(other, attr)
                else:
                    obs[other.id] = self.null_value
            return {self.channel: obs}
        else:
            return {}

    @abstractmethod
    def get_obs(self, agent, **kwargs):
        pass

    @abstractproperty
    def channel(self):
        pass

    @abstractproperty
    def null_value(self):
        pass


# --------------------- #
# --- Communication --- #
# --------------------- #

class BroadcastObserver(Observer):
    """
    Observe the broadcast state of broadcasting agents.
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._set_obs_space_simple(
            BroadcastObservingAgent, lambda *args: Box(-1, 1, (1,)), **kwargs
        )

    def get_obs(self, agent, **kwargs):
        return self._get_obs(
            agent,
            instance=BroadcastObservingAgent,
            other_instance=BroadcastingAgent,
            attr='broadcasting',
            **kwargs
        )

    @property
    def channel(self):
        return 'broadcast'

    @property
    def null_value(self):
        return -1


# ----------------------- #
# --- Health and Life --- #
# ----------------------- #

class HealthObserver(Observer):
    """
    Observe the health state of all the agents in the simulator.
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._set_obs_space_simple(
            HealthObservingAgent, lambda other: Box(-1, other.max_health, (1,)), **kwargs
        )

    def get_obs(self, agent, **kwargs):
        """
        Get the health state of all the agents in the simulator.

        agent (HealthObservingAgent):
            The agent making the observation.
        """
        return self._get_obs(
            agent,
            instance=HealthObservingAgent,
            attr='health',
            **kwargs
        )

    @property
    def channel(self):
        return 'health'

    @property
    def null_value(self):
        return -1


class LifeObserver(Observer):
    """
    Observe the life state of all the agents in the simulator.
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._set_obs_space_simple(
            LifeObservingAgent, lambda *args: Box(-1, 1, (1,), np.int), **kwargs
        )

    def get_obs(self, agent, **kwargs):
        """
        Get the life state of all the agents in the simulator.

        agent (LifeObservingAgent):
            The agent making the observation.
        """
        return self._get_obs(
            agent,
            instance=LifeObservingAgent,
            attr='is_alive',
            **kwargs
        )

    @property
    def channel(self):
        return 'life'

    @property
    def null_value(self):
        return -1


# ----------------------------- #
# --- Position and Movement --- #
# ----------------------------- #

class PositionObserver(Observer):
    """
    Observe the positions of all the agents in the simulator.
    """
    def __init__(self, position_state=None, **kwargs):
        super().__init__(**kwargs)
        self.position_state = position_state
        self._set_obs_space_simple(
            PositionObservingAgent,
            lambda *args: Box(-1, self.position_state.region, (2,), np.int), **kwargs
        )

    def get_obs(self, agent, **kwargs):
        """
        Get the positions of all the agents in the simulator.
        """
        return self._get_obs(agent, instance=PositionObservingAgent, attr='position')

    @property
    def channel(self):
        return 'position'

    @property
    def null_value(self):
        return np.array([-1, -1])


class RelativePositionObserver(Observer):
    """
    Observe the relative positions of agents in the simulator.
    """
    def __init__(self, position_state=None, **kwargs):
        super().__init__(**kwargs)
        self.position_state = position_state
        self._set_obs_space_simple(
            PositionObservingAgent,
            lambda *args: Box(
                -self.position_state.region, self.position_state.region, (2,), np.int
            ), **kwargs
        )

    def get_obs(self, agent, **kwargs):
        """
        Get the relative positions of all the agents in the simulator.
        """
        if isinstance(agent, PositionObservingAgent):
            obs = {}
            for other in self.agents.values():
                r_diff = other.position[0] - agent.position[0]
                c_diff = other.position[1] - agent.position[1]
                obs[other.id] = np.array([r_diff, c_diff])
            return {self.channel: obs}
        else:
            return {}

    @property
    def channel(self):
        return 'relative_position'

    @property
    def null_value(self):
        return np.array([-self.position_state.region, -self.position_state.region])


class GridPositionBasedObserver:
    """
    Agents observe a grid of size agent_view centered on their
    position. The values of the cells are as such:
        Out of bounds  : -1
        Empty          :  0
        Agent occupied : 1

    position (PositionState):
        The position state handler, which contains the region.

    agents (dict):
        The dictionary of agents.
    """
    def __init__(self, position_state=None, agents=None, **kwargs):
        self.position_state = position_state
        self.agents = agents

        for agent in agents.values():
            if isinstance(agent, AgentObservingAgent) and \
               isinstance(agent, PositionObservingAgent):
                agent.observation_space['position'] = Box(
                    -1, 1, (agent.agent_view*2+1, agent.agent_view*2+1), np.int
                )

    def get_obs(self, my_agent, **kwargs):
        """
        Generate an observation of other agents in the grid surrounding this agent's
        position.
        """
        if isinstance(my_agent, AgentObservingAgent) and \
           isinstance(my_agent, PositionObservingAgent):
            signal = np.zeros((my_agent.agent_view*2+1, my_agent.agent_view*2+1))

            # --- Determine the boundaries of the agents' grids --- #
            # For left and top, we just do: view - x,y >= 0
            # For the right and bottom, we just do region - x,y - 1 - view > 0
            if my_agent.agent_view - my_agent.position[0] >= 0: # Top end
                signal[0:my_agent.agent_view - my_agent.position[0], :] = -1
            if my_agent.agent_view - my_agent.position[1] >= 0: # Left end
                signal[:, 0:my_agent.agent_view - my_agent.position[1]] = -1
            if self.position_state.region - my_agent.position[0] - my_agent.agent_view - 1 < 0:
                # Bottom end
                signal[
                    self.position_state.region - my_agent.position[0] - my_agent.agent_view - 1:,
                    :
                ] = -1
            if self.position_state.region - my_agent.position[1] - my_agent.agent_view - 1 < 0:
                # Right end
                signal[
                    :, self.position_state.region - my_agent.position[1] - my_agent.agent_view - 1:
                ] = -1

            # --- Determine the positions of all the other alive agents --- #
            for other_id, other_agent in self.agents.items():
                if other_id == my_agent.id: continue # Don't observe yourself
                if not other_agent.is_alive: continue # Can only observe alive agents
                r_diff = other_agent.position[0] - my_agent.position[0]
                c_diff = other_agent.position[1] - my_agent.position[1]
                if -my_agent.agent_view <= r_diff <= my_agent.agent_view and \
                        -my_agent.agent_view <= c_diff <= my_agent.agent_view:
                    r_diff += my_agent.agent_view
                    c_diff += my_agent.agent_view
                    signal[r_diff, c_diff] = 1 # There is an agent at this location.

            return {'position': signal}
        else:
            return {}


class GridPositionTeamBasedObserver:
    """
    Agents observe a grid of size agent_view centered on their
    position. The observation contains one channel per team, where the value of
    the cell is the number of agents on that team that occupy that square. -1
    indicates out of bounds.

    position (PositionState):
        The position state handler, which contains the region.

    number_of_teams (int):
        The number of teams in this simuation.
        Default 0.

    agents (dict):
        The dictionary of agents.
    """
    def __init__(self, position_state=None, number_of_teams=0, agents=None, **kwargs):
        self.position_state = position_state
        self.number_of_teams = number_of_teams + 1
        self.agents = agents

        for agent in self.agents.values():
            if isinstance(agent, AgentObservingAgent) and \
               isinstance(agent, PositionObservingAgent):
                agent.observation_space['position'] = Box(
                    -1, np.inf, (agent.agent_view*2+1, agent.agent_view*2+1, self.number_of_teams),
                    np.int
                )

    def get_obs(self, my_agent, **kwargs):
        """
        Generate an observation of other agents in the grid surrounding this agent's
        position. Each team has its own channel and the value represents the number
        of agents of that team occupying the same square.
        """
        if isinstance(my_agent, AgentObservingAgent) and \
           isinstance(my_agent, PositionObservingAgent):
            signal = np.zeros((my_agent.agent_view*2+1, my_agent.agent_view*2+1))

            # --- Determine the boundaries of the agents' grids --- #
            # For left and top, we just do: view - x,y >= 0
            # For the right and bottom, we just do region - x,y - 1 - view > 0
            if my_agent.agent_view - my_agent.position[0] >= 0: # Top end
                signal[0:my_agent.agent_view - my_agent.position[0], :] = -1
            if my_agent.agent_view - my_agent.position[1] >= 0: # Left end
                signal[:, 0:my_agent.agent_view - my_agent.position[1]] = -1
            if self.position_state.region - my_agent.position[0] - my_agent.agent_view - 1 < 0:
                # Bottom end
                signal[
                    self.position_state.region - my_agent.position[0] - my_agent.agent_view - 1:,
                    :
                ] = -1
            if self.position_state.region - my_agent.position[1] - my_agent.agent_view - 1 < 0:
                # Right end
                signal[
                    :,
                    self.position_state.region - my_agent.position[1] - my_agent.agent_view - 1:
                ] = -1

            # Repeat the boundaries signal for all teams
            signal = np.repeat(signal[:, :, np.newaxis], self.number_of_teams, axis=2)

            # --- Determine the positions of all the other alive agents --- #
            for other_id, other_agent in self.agents.items():
                if other_id == my_agent.id: continue # Don't observe yourself
                if not other_agent.is_alive: continue # Can only observe alive agents
                r_diff = other_agent.position[0] - my_agent.position[0]
                c_diff = other_agent.position[1] - my_agent.position[1]
                if -my_agent.agent_view <= r_diff <= my_agent.agent_view and \
                        -my_agent.agent_view <= c_diff <= my_agent.agent_view:
                    r_diff += my_agent.agent_view
                    c_diff += my_agent.agent_view
                    signal[r_diff, c_diff, other_agent.team] += 1

            return {'position': signal}
        else:
            return {}


class SpeedObserver(Observer):
    """
    Observe the speed of all the agents in the simulator.
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._set_obs_space(
            SpeedAngleObservingAgent,
            SpeedAngleAgent,
            lambda other: Box(-1, other.max_speed, (1,)),
            lambda: Box(-1, -1, (1,)), **kwargs
        )

    def get_obs(self, agent, **kwargs):
        """
        Get the speed of all the agents in the simulator.
        """
        return self._get_obs(
            agent,
            instance=SpeedAngleObservingAgent,
            other_instance=SpeedAngleAgent,
            attr='speed',
            **kwargs
        )

    @property
    def channel(self):
        return 'speed'

    @property
    def null_value(self):
        return -1


class AngleObserver(Observer):
    """
    Observe the angle of all the agents in the simulator.
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._set_obs_space(
            SpeedAngleObservingAgent,
            SpeedAngleAgent,
            lambda *args: Box(-1, 360, (1,)),
            lambda *args: Box(-1, -1, (1,)), **kwargs
        )

    def get_obs(self, agent, **kwargs):
        """
        Get the angle of all the agents in the simulator.
        """
        return self._get_obs(
            agent,
            instance=SpeedAngleObservingAgent,
            other_instance=SpeedAngleAgent,
            attr='ground_angle',
            **kwargs
        )

    @property
    def channel(self):
        return 'ground_angle'

    @property
    def null_value(self):
        return -1


class VelocityObserver(Observer):
    """
    Observe the velocity of all the agents in the simulator.
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._set_obs_space(
            VelocityObservingAgent, VelocityAgent,
            lambda other: Box(-other.max_speed, other.max_speed, (2,)),
            lambda: Box(0, 0, (2,)), **kwargs
        )

    def get_obs(self, agent, **kwargs):
        """
        Get the velocity of all the agents in the simulator.
        """
        return self._get_obs(
            agent,
            instance=VelocityObservingAgent,
            other_instance=VelocityAgent,
            attr='velocity',
            **kwargs
        )

    @property
    def channel(self):
        return 'velocity'

    @property
    def null_value(self):
        return np.zeros(2)


# -------------------------------- #
# --- Resources and Harvesting --- #
# -------------------------------- #

class GridResourceObserver:
    """
    Agents observe a grid of size resource_view centered on their
    position. The values in the grid are the values of the resources in that
    area.

    resources (ResourceState):
        The resource state handler.

    agents (dict):
        The dictionary of agents.
    """
    def __init__(self, resource_state=None, agents=None, **kwargs):
        self.resource_state = resource_state
        self.agents = agents

        for agent in agents.values():
            if isinstance(agent, ResourceObservingAgent):
                agent.observation_space['resources'] = Box(
                    -1, self.resource_state.max_value,
                    (agent.resource_view*2+1, agent.resource_view*2+1)
                )

    def get_obs(self, agent, **kwargs):
        """
        These cells are filled with the values of the resources surrounding the
        agent's position.
        """
        if isinstance(agent, ResourceObservingAgent):
            signal = -np.ones((agent.resource_view*2+1, agent.resource_view*2+1))

            # Derived by considering each square in the resources as an "agent" and
            # then applied the agent diff logic from above. The resulting for-loop
            # can be written in the below vectorized form.
            (r, c) = agent.position
            r_lower = max([0, r-agent.resource_view])
            r_upper = min([self.resource_state.region-1, r+agent.resource_view])+1
            c_lower = max([0, c-agent.resource_view])
            c_upper = min([self.resource_state.region-1, c+agent.resource_view])+1
            signal[
                (r_lower+agent.resource_view-r):(r_upper+agent.resource_view-r),
                (c_lower+agent.resource_view-c):(c_upper+agent.resource_view-c)
            ] = self.resource_state.resources[r_lower:r_upper, c_lower:c_upper]
            return {'resources': signal}
        else:
            return {}


# ------------ #
# --- Team --- #
# ------------ #

class TeamObserver(Observer):
    """
    Observe the team of each agent in the simulator.
    """
    def __init__(self, number_of_teams=0, **kwargs):
        super().__init__(**kwargs)
        self.number_of_teams = number_of_teams
        self._set_obs_space_simple(
            TeamObservingAgent, lambda *args: Box(-1, self.number_of_teams, (1,), np.int), **kwargs
        )

    def get_obs(self, agent, **kwargs):
        """
        Get the team of each agent in the simulator.
        """
        return self._get_obs(
            agent,
            instance=TeamObservingAgent,
            attr='team',
            **kwargs
        )

    @property
    def channel(self):
        return 'team'

    @property
    def null_value(self):
        return -1



from abc import ABC, abstractmethod, abstractproperty

from gym.spaces import Discrete, Box
import numpy as np

from abmarl.sim.components.agent import AttackingAgent, GridMovementAgent, HarvestingAgent, \
    SpeedAngleAgent, AcceleratingAgent, VelocityAgent, \
    CollisionAgent, BroadcastingAgent


class Actor(ABC):
    """
    Base actor class provides the interface required of all actors. Setup the agents'
    action space according to the Actor's channel.

        agents (dict):
            The dictionary of agents.

        instance (Agent):
            An Agent class. This is used in the isinstance check to determine if
            the agent will receive the action channel.

        space_func (function):
            A function that takes the agent as input and outputs the action space.
    """
    def __init__(self, agents=None, instance=None, space_func=None, **kwargs):
        self.agents = agents
        for agent in self.agents.values():
            if isinstance(agent, instance):
                agent.action_space[self.channel] = space_func(agent)

    def _get_action_from_dict(self, action_dict, **kwargs):
        """
        The action dict passed to an AgentBasedSimulation will be keyed off the
        agent id's. This action dict, however, is just the specific agent's actions,
        which is a dictionary keyed off the action channels. This function extracts
        that action if available, otherwise returning the Actor's null_value.
        """
        return action_dict.get(self.channel, self.null_value)

    @abstractmethod
    def process_action(self, agent, action_dict, **kwargs):
        pass

    @abstractproperty
    def channel(self):
        pass

    @abstractproperty
    def null_value(self):
        pass


# ----------------- #
# --- Attacking --- #
# ----------------- #

class AttackActor(Actor):
    """
    Agents can attack other agents within their attack radius. If there are multiple
    attackable agents in the radius, then one will be randomly chosen. Attackable
    agents are determiend by the team_matrix.

    agents (dict of Agents):
        The dictionary of agents.

    attack_norm (int):
        Norm used to measure the distance between agents. For example, you might
        use a norm of 1 or np.inf in a Gird space, while 2 might be used in a Continuous
        space. Default is np.inf.

    team_attack_matrix (np.ndarray):
        A matrix that indicates which teams can attack which other team using the
        value at the index, like so:
            team_matrix[attacking_team, attacked_team] = 0 or 1.
        0 indicates it cannot attack, 1 indicates that it can.
        Default None, meaning that any team can attack any other team, and no team
        can attack itself.

    number_of_teams (int):
        Specify the number of teams in the simulation for building the team_attack_matrix
        if that is not specified here.
        Default 0, indicating that there are no teams and its a free-for-all battle.
    """
    def __init__(self, attack_norm=np.inf, team_attack_matrix=None, number_of_teams=0, **kwargs):
        super().__init__(
            instance=AttackingAgent,
            space_func=lambda agent: Discrete(2),
            **kwargs
        )
        if team_attack_matrix is None:
            # Default: teams can attack all other teams but not themselves. Agents
            # that are "on team 0" are actually teamless, so they can be attacked
            # by and can attack agents from any other team, including "team 0"
            # agents.
            self.team_attack_matrix = -np.diag(np.ones(number_of_teams+1)) + 1
            self.team_attack_matrix[0, 0] = 1
        else:
            self.team_attack_matrix = team_attack_matrix
        self.attack_norm = attack_norm

    def process_action(self, attacking_agent, action_dict, **kwargs):
        """
        If the agent has chosen to attack, then determine which agent got attacked.

        return (Agent):
            Return the attacked agent object. This can be None if no agent was
            attacked.
        """
        if self._get_action_from_dict(action_dict):
            for attacked_agent in self.agents.values():
                if attacked_agent.id == attacking_agent.id:
                    # Cannot attack yourself
                    continue
                elif not attacked_agent.is_alive:
                    # Cannot attack a dead agent
                    continue
                elif np.linalg.norm(attacking_agent.position - attacked_agent.position,
                                    self.attack_norm) > attacking_agent.attack_range:
                    # Agent is too far away
                    continue
                elif not self.team_attack_matrix[attacking_agent.team, attacked_agent.team]:
                    # Attacking agent cannot attack this agent
                    continue
                elif np.random.uniform() > attacking_agent.attack_accuracy:
                    # Attempted attack, but it failed
                    continue
                else:
                    # The agent was successfully attacked!
                    return attacked_agent

    @property
    def channel(self):
        return 'attack'

    @property
    def null_value(self):
        return False


# --------------------- #
# --- Communication --- #
# --------------------- #

class BroadcastActor(Actor):
    """
    BroadcastingAgents can choose to broadcast in this step or not.

    broadcast_state (BroadcastState):
        The broadcast state handler. Needed to modifying the agents' broadcasting state.

    agents (dict):
        Dictionary of agents.
    """
    def __init__(self, broadcast_state=None, **kwargs):
        super().__init__(
            instance=BroadcastingAgent,
            space_func=lambda agent: Discrete(2),
            **kwargs
        )
        self.broadcast_state = broadcast_state

    def process_action(self, agent, action_dict, **kwargs):
        """
        Determine the agents new broadcasting state based on its action.

        return: bool
            The agent's broadcasting state.
        """
        broadcasting = self._get_action_from_dict(action_dict)
        self.broadcast_state.modify_broadcast(agent, broadcasting)

    @property
    def channel(self):
        return 'broadcast'

    @property
    def null_value(self):
        return False


# ----------------------------- #
# --- Position and Movement --- #
# ----------------------------- #

class GridMovementActor(Actor):
    """
    Provides the necessary action space for agents who can move and processes such
    movements.

    position (GridPositionState):
        The position state handler. Needed to modify the agents' positions.

    agents (dict):
        The dictionary of agents.
    """
    def __init__(self, position_state=None, **kwargs):
        super().__init__(
            instance=GridMovementAgent,
            space_func=lambda agent: Box(-agent.move_range, agent.move_range, (2,), np.int),
            **kwargs
        )
        self.position_state = position_state

    def process_action(self, agent, action_dict, **kwargs):
        """
        Determine the agent's new position based on its move action.

        return (np.array):
            How much the agent has moved in row and column. This can be different
            from the desired move if the position update was invalid.
        """
        move = self._get_action_from_dict(action_dict)
        position_before = agent.position
        self.position_state.modify_position(agent, move, **kwargs)
        return agent.position - position_before

    @property
    def channel(self):
        return 'move'

    @property
    def null_value(self):
        return np.zeros(2)


class SpeedAngleMovementActor:
    """
    Process acceleration and angle changes for SpeedAngleAgents. Update the agents'
    positions based on their new speed and direction.

    position (ContinuousPositionState):
        The position state handler. Needed to modify agent positions.

    speed_angle (SpeedAngleState):
        The speed and angle state handler.

    agents (dict):
        The dictionary of agents.
    """
    def __init__(self, position_state=None, speed_angle_state=None, agents=None, **kwargs):
        self.position_state = position_state
        self.speed_angle_state = speed_angle_state
        self.agents = agents

        for agent in agents.values():
            if isinstance(agent, SpeedAngleAgent):
                agent.action_space['accelerate'] = Box(
                    -agent.max_acceleration, agent.max_acceleration, (1,)
                )
                agent.action_space['bank'] = Box(
                    -agent.max_banking_angle_change, agent.max_banking_angle_change, (1,)
                )

    def process_move(self, agent, acceleration, angle, **kwargs):
        """
        Update the agent's speed by applying the acceleration and the agent's banking
        angle by applying the change. Then use the updated speed and ground angle
        to determine the agent's next position.

        agent (SpeedAngleAgent):
            Agent that is attempting to move.

        acceleration (np.array):
            A one-element float array that changes the agent's speed. New speed
            must be within the agent's min and max speed.

        angle (np.array):
            A one-element float array that changes the agent's banking angle. New
            banking angle must be within the agent's min and max banking angles.

        return (np.array):
            Return the change in position.
        """
        if isinstance(agent, SpeedAngleAgent):
            self.speed_angle_state.modify_speed(agent, acceleration[0])
            self.speed_angle_state.modify_banking_angle(agent, angle[0])

            x_position = agent.speed*np.cos(np.deg2rad(agent.ground_angle))
            y_position = agent.speed*np.sin(np.deg2rad(agent.ground_angle))

            position_before = agent.position
            self.position_state.modify_position(agent, np.array([x_position, y_position]))
            return agent.position - position_before


class AccelerationMovementActor(Actor):
    """
    Process x,y accelerations for AcceleratingAgents, which are given an 'accelerate'
    action. Update the agents' positions based on their new velocity.

    position_state (ContinuousPositionState):
        The position state handler. Needed to modify agent positions.

    velocity_state (VelocityState):
        The velocity state handler. Needed to modify agent velocities.

    agents (dict):
        The dictionary of agents.
    """
    def __init__(self, position_state=None, velocity_state=None, **kwargs):
        super().__init__(
            instance=AcceleratingAgent,
            space_func=lambda agent: Box(-agent.max_acceleration, agent.max_acceleration, (2,)),
            **kwargs
        )
        self.position_state = position_state
        self.velocity_state = velocity_state

    def process_action(self, agent, action_dict, **kwargs):
        """
        Update the agent's velocity by applying the acceleration. Then use the
        updated velocity to determine the agent's next position.

        return (np.array):
            Return the change in position.
        """
        acceleration = self._get_action_from_dict(action_dict)
        self.velocity_state.modify_velocity(agent, acceleration)
        position_before = agent.position
        self.position_state.modify_position(agent, agent.velocity, **kwargs)
        return agent.position - position_before

    @property
    def channel(self):
        return 'accelerate'

    @property
    def null_value(self):
        return np.zeros(2)


# -------------------------------- #
# --- Resources and Harvesting --- #
# -------------------------------- #

class GridResourcesActor(Actor):
    """
    Provides the necessary action space for agents who can harvest resources and
    processes the harvesting action.

    resources (ResourceState):
        The resource state handler.

    agents (dict):
        The dictionary of agents.
    """
    def __init__(self, resource_state=None, **kwargs):
        super().__init__(
            instance=HarvestingAgent,
            space_func=lambda agent: Box(0, agent.max_harvest, (1,)),
            **kwargs
        )
        self.resource_state = resource_state

    def process_action(self, agent, action_dict, **kwargs):
        """
        Harvest some amount of resources at the agent's position.

        return (float):
            Return the amount of resources that was actually harvested. This can
            be less than the desired amount if the cell does not have enough resources.
        """
        amount = self._get_action_from_dict(action_dict)
        location = tuple(agent.position)
        resource_before = self.resource_state.resources[location]
        self.resource_state.modify_resources(location, -amount)
        return resource_before - self.resource_state.resources[location]

    @property
    def channel(self):
        return 'harvest'

    @property
    def null_value(self):
        return 0


# --------------------------------------------- #
# --- Actors that don't receive agent input --- #
# --------------------------------------------- #

class ContinuousCollisionActor:
    """
    Identify collisions among agents and update positions and velocities according to
    elastic collision physics.

    position_state (PositionState):
        The PositionState handler.

    velocity_state (VelocityState):
        The VelocityState handler.

    agents (dict):
        The dictionary of agents.
    """
    def __init__(self, position_state=None, velocity_state=None, agents=None, **kwargs):
        self.position_state = position_state
        self.velocity_state = velocity_state
        self.agents = agents

    def detect_collisions_and_modify_states(self, **kwargs):
        """
        Detect collisions between agents and update position and velocities.
        """
        checked_agents = set()
        for agent1 in self.agents.values():
            if not (isinstance(agent1, CollisionAgent) and isinstance(agent1, VelocityAgent)):
                continue
            checked_agents.add(agent1.id)
            for agent2 in self.agents.values():
                if not (isinstance(agent1, VelocityAgent) and isinstance(agent2, CollisionAgent)):
                    continue
                if agent1.id == agent2.id: continue # Cannot collide with yourself
                if agent2.id in checked_agents: continue # Already checked this agent
                dist = np.linalg.norm(agent1.position - agent2.position)
                combined_sizes = agent1.size + agent2.size
                if dist < combined_sizes:
                    self._undo_overlap(agent1, agent2, dist, combined_sizes)
                    self._update_velocities(agent1, agent2)

    def _undo_overlap(self, agent1, agent2, dist, combined_sizes, **kwargs):
        """
        Colliding agents can overlap within a timestep. So we need to move the
        colliding agents "backwards" through their path in order to capture the
        positions they were in when they actually collided.

        agent1 (CollisionAgent):
            One of the colliding agents.

        agent2 (CollisionAgent):
            The other colliding agent.

        dist (float):
            The collision distance threshold.

        combined_size (float):
            The combined size of the two agents
        """
        overlap = (combined_sizes - dist) / combined_sizes
        self.position_state.modify_position(agent1, -agent1.velocity * overlap)
        self.position_state.modify_position(agent2, -agent2.velocity * overlap)

    def _update_velocities(self, agent1, agent2, **kwargs):
        """
        Updates the velocities of two agents when they collide based on an
        elastic collision assumption.

        agent1 (CollisionAgent):
            One of the colliding agents.

        agent2 (CollisionAgent):
            The other colliding agent.
        """
        # calculate vector between centers
        rel_position = [
            agent2.position - agent1.position,
            agent1.position - agent2.position
        ]
        # Calculate relative velocities
        rel_velocities = [
            agent1.velocity - agent2.velocity,
            agent2.velocity - agent1.velocity
        ]
        # Calculate mass factor
        mass_factor = [
            2 * agent2.mass / (agent2.mass + agent1.mass),
            2 * agent1.mass / (agent2.mass + agent1.mass)
        ]
        # norm
        norm = [
            np.square(np.linalg.norm(rel_position[0])),
            np.square(np.linalg.norm(rel_position[1]))
        ]
        # Dot product of relative velocity and relative distcance
        dot = [
            np.dot(rel_velocities[0], rel_position[0]),
            np.dot(rel_velocities[1], rel_position[1])
        ]
        # bringing it all together
        vel_new = [
            agent1.velocity - (mass_factor[0] * (dot[0]/norm[0]) * rel_position[0]),
            agent2.velocity - (mass_factor[1] * (dot[1]/norm[1]) * rel_position[1])
        ]
        # Only update the velocity if not stationary
        self.velocity_state.set_velocity(agent1, vel_new[0])
        self.velocity_state.set_velocity(agent2, vel_new[1])



import numpy as np

from abmarl.sim import PrincipleAgent, ActingAgent, ObservingAgent


# ------------------ #
# --- Base Agent --- #
# ------------------ #

class ComponentAgent(PrincipleAgent):
    """
    Component Agents have a position, life, and team.

    initial_position (np.array or None):
        The desired starting position for this agent.

    min_health (float):
        The minimum health the agent can reach before it dies.
        Default 0.0.

    max_health (float):
        The maximum health the agent can reach.
        Default 1.0.

    initial_health (float or None):
        The initial health of the agent. The health will be set to this initial
        option at reset time.

    team (int or None):
        The agent's team. Teams are indexed starting from 1, with team 0 reserved
        for agents that are not on a team (None).
    """
    def __init__(self, initial_position=None, min_health=0.0, max_health=1.0, initial_health=None,
                 team=None, **kwargs):
        super().__init__(**kwargs)
        self.initial_position = initial_position
        assert type(min_health) in [float, int] and type(max_health) in [float, int], \
            "Min and max health must be numerical."
        assert min_health <= max_health, \
            "The min health must be less than or equal to the max_health."
        self._min_max_health = np.array([min_health, max_health])
        self.initial_health = initial_health
        self.is_alive = True
        self.team = team

    @property
    def initial_position(self):
        return self._initial_position

    @initial_position.setter
    def initial_position(self, value):
        if value is not None:
            assert type(value) is np.ndarray, "Initial position must be a numpy array."
            assert value.shape == (2,), "Initial position must be a 2-dimensional array."
            assert value.dtype in [np.int, np.float], "Initial position must be numerical."
        self._initial_position = value

    @property
    def min_health(self):
        return self._min_max_health[0]

    @property
    def max_health(self):
        return self._min_max_health[1]

    @property
    def initial_health(self):
        return self._initial_health

    @initial_health.setter
    def initial_health(self, value):
        if value is not None:
            assert type(value) in [float, int], "Initial health must be a float."
            assert self.min_health <= value <= self.max_health, \
                "Initial health must be between the min and max health."
        self._initial_health = value

    @property
    def team(self):
        return self._team

    @team.setter
    def team(self, value):
        if value is not None:
            assert type(value) is int, "Team must be an int."
            assert value != 0, "Team 0 is reserved for agents who do not have a team. " + \
                "Use a team number greater than 0."
            self._team = value
        else:
            self._team = 0

    @property
    def configured(self):
        """
        Determine if the agent has been successfully configured.
        """
        return super().configured and \
            self._min_max_health is not None and \
            self.is_alive is not None and \
            self.team is not None


# ----------------- #
# --- Attacking --- #
# ----------------- #

class AttackingAgent(ActingAgent, ComponentAgent):
    """
    Agents that can attack other agents.

    attack_range (int):
        The effective range of the attack. Can be used to determine if an attack
        is successful based on distance between agents.

    attack_strength (float):
        How effective the agent's attack is. This is applicable in situations where
        the agents' health is affected by attacks.

    attack_accuracy (float):
        The effective accuracy of the agent's attack. Should be between 0 and 1.
        To make deterministic attacks, use 1. Default is 1.
    """
    def __init__(self, attack_range=None, attack_strength=None, attack_accuracy=1, **kwargs):
        super().__init__(**kwargs)
        assert attack_range is not None, "attack_range must be a nonnegative integer"
        self.attack_range = attack_range
        assert attack_strength is not None, "attack_strength must be a nonnegative number"
        self.attack_strength = attack_strength
        self.attack_accuracy = attack_accuracy

    @property
    def configured(self):
        """
        The agent is successfully configured if the attack range and strength is
        specified.
        """
        return super().configured and self.attack_range is not None and \
            self.attack_strength is not None


# --------------------- #
# --- Communication --- #
# --------------------- #

class BroadcastingAgent(ActingAgent, ComponentAgent):
    """
    BroadcastingAgents can broadcast their observation within some range of their
    position.

    braodcast_range (int):
        The agent's broadcasting range.
    """
    def __init__(self, broadcast_range=None, **kwargs):
        super().__init__(**kwargs)
        self.broadcast_range = broadcast_range
        self.broadcasting = False

    @property
    def configured(self):
        """
        The agent is successfully configured if the broadcast range is specified.
        """
        return super().configured and self.broadcast_range is not None


class BroadcastObservingAgent(ObservingAgent, ComponentAgent): pass


# ----------------------- #
# --- Health and Life --- #
# ----------------------- #

class LifeObservingAgent(ObservingAgent, ComponentAgent): pass
class HealthObservingAgent(ObservingAgent, ComponentAgent): pass


# ----------------- #
# --- Observing --- #
# ----------------- #

# TODO: move this to a more specific location
class AgentObservingAgent(ObservingAgent, ComponentAgent):
    """
    Agents can observe other agents.

    agent_view (int):
        Any agent within this many spaces will be fully observed.
    """
    def __init__(self, agent_view=None, **kwargs):
        """
        Agents can see other agents up to some maximal distance away, indicated
        by the view.
        """
        super().__init__(**kwargs)
        assert agent_view is not None, "agent_view must be nonnegative integer"
        self.agent_view = agent_view

    @property
    def configured(self):
        """
        Agents are configured if the agent_view parameter is set.
        """
        return super().configured and self.agent_view is not None


# ----------------------------- #
# --- Position and Movement --- #
# ----------------------------- #


class PositionObservingAgent(ObservingAgent, ComponentAgent): pass


class GridMovementAgent(ActingAgent, ComponentAgent):
    """
    Agents can move up to some number of spaces away.

    move_range (int):
        The maximum number of cells away that the agent can move.
    """
    def __init__(self, move_range=None, **kwargs):
        super().__init__(**kwargs)
        assert move_range is not None, "move_range must be an integer"
        self.move_range = move_range

    @property
    def configured(self):
        """
        Agents are configured if the move_range parameter is set.
        """
        return super().configured and self.move_range is not None


class SpeedAngleAgent(ComponentAgent):
    """
    Agents have a speed and a banking angle which are used to determine how the
    agent moves around a continuous field.

    min_speed (float):
        The minimum speed this agent can travel.

    max_speed (float):
        The maximum speed this agent can travel.

    max_banking_angle (float):
        The maximum banking angle the agent can endure.

    initial_speed (float):
        The agent's initial speed.

    initial_banking_angle (float):
        The agent's initial banking angle.

    initial_ground_angle (float):
        The agent's initial ground angle.
    """
    def __init__(self, min_speed=0.25, max_speed=1.0, max_banking_angle=45, initial_speed=None,
                 initial_banking_angle=None, initial_ground_angle=None, **kwargs):
        super().__init__(**kwargs)
        self.min_speed = min_speed
        self.max_speed = max_speed
        self.initial_speed = initial_speed
        self.speed = None # Should be set by the state handler

        self.max_banking_angle = max_banking_angle
        self.initial_banking_angle = initial_banking_angle
        self.initial_ground_angle = initial_ground_angle
        self.banking_angle = None # Should be set by the state handler

    @property
    def configured(self):
        return super().configured and self.min_speed is not None and self.max_speed is not None \
            and self.max_banking_angle is not None


class SpeedAngleActingAgent(ActingAgent, ComponentAgent):
    """
    Agents can change their speed and banking angles.

    max_acceleration (float):
        The maximum amount by which an agent can change its speed in a single time
        step.

    max_banking_angle_change (float):
        The maximum amount by which an agent can change its banking angle in a
        single time step.
    """
    def __init__(self, max_acceleration=0.25, max_banking_angle_change=30, **kwargs):
        super().__init__(**kwargs)
        self.max_acceleration = max_acceleration
        self.max_banking_angle_change = max_banking_angle_change

    @property
    def configured(self):
        return super().configured and self.max_acceleration is not None and \
            self.max_banking_angle_change is not None


class SpeedAngleObservingAgent(ObservingAgent, ComponentAgent): pass


class VelocityAgent(ComponentAgent):
    """
    Agents have a velocity which determines how it moves around in a continuous
    field. Agents can accelerate to modify their velocities.

    initial_velocity (np.array):
        Two-element float array that is the agent's initial velocity.

    max_speed (float):
        The maximum speed the agent can travel.

    max_acceleration (float):
        The maximum amount by which an agent can change its velocity in a single
        time step.
    """
    def __init__(self, initial_velocity=None, max_speed=None, **kwargs):
        super().__init__(**kwargs)
        self.initial_velocity = initial_velocity
        self.max_speed = max_speed

    @property
    def configured(self):
        return super().configured and self.max_speed is not None


class AcceleratingAgent(ActingAgent, ComponentAgent):
    """
    Agents can accelerate to modify their velocities.

    max_acceleration (float):
        The maximum amount by which an agent can change its velocity in a single
        time step.
    """
    def __init__(self, max_acceleration=None, **kwargs):
        super().__init__(**kwargs)
        self.max_acceleration = max_acceleration

    @property
    def configured(self):
        return super().configured and self.max_acceleration is not None


class VelocityObservingAgent(ObservingAgent, ComponentAgent): pass


class CollisionAgent(PrincipleAgent):
    """
    Agents that have physical size and mass and can be used in collisions.

    size (float):
        The size of the agent.
        Default 1.

    mass (float):
        The mass of the agent.
        Default 1.
    """
    def __init__(self, size=1, mass=1, **kwargs):
        super().__init__(**kwargs)
        self.size = size
        self.mass = mass

    @property
    def configured(self):
        return super().configured and self.size is not None and self.mass is not None


# -------------------------------- #
# --- Resources and Harvesting --- #
# -------------------------------- #

class HarvestingAgent(ActingAgent, ComponentAgent):
    """
    Agents can harvest resources.

    max_harvest (double):
        The maximum amount of resources the agent can harvest from the cell it
        occupies.
    """
    def __init__(self, max_harvest=None, **kwargs):
        super().__init__(**kwargs)
        assert max_harvest is not None, "max_harvest must be nonnegative number"
        self.max_harvest = max_harvest

    @property
    def configured(self):
        """
        Agents are configured if max_harvest is set.
        """
        return super().configured and self.max_harvest is not None


class ResourceObservingAgent(ObservingAgent, ComponentAgent):
    """
    Agents can observe the resources in the simulation.

    resource_view (int):
        Any resources within this range of the agent's position will be fully observed.
    """
    def __init__(self, resource_view=None, **kwargs):
        super().__init__(**kwargs)
        assert resource_view is not None, "resource_view must be nonnegative integer"
        self.resource_view = resource_view

    @property
    def configured(self):
        """
        Agents are configured if the resource_view parameter is set.
        """
        return super().configured and self.resource_view is not None


# ------------ #
# --- Team --- #
# ------------ #

class TeamObservingAgent(ObservingAgent, ComponentAgent): pass



import numpy as np

from abmarl.sim.components.state import GridPositionState, ContinuousPositionState, \
    SpeedAngleState, VelocityState
from abmarl.sim.components.actor import GridMovementActor, SpeedAngleMovementActor, \
    AccelerationMovementActor
from abmarl.sim.components.agent import SpeedAngleAgent, SpeedAngleActingAgent, VelocityAgent, \
    GridMovementAgent, AcceleratingAgent


class GridMovementTestAgent(GridMovementAgent): pass


def test_grid_movement_component():
    agents = {
        'agent0': GridMovementTestAgent(
            id='agent0', initial_position=np.array([6, 4]), move_range=2
        ),
        'agent1': GridMovementTestAgent(
            id='agent1', initial_position=np.array([3, 3]), move_range=3
        ),
        'agent2': GridMovementTestAgent(
            id='agent2', initial_position=np.array([0, 1]), move_range=1
        ),
        'agent3': GridMovementTestAgent(
            id='agent3', initial_position=np.array([8, 4]), move_range=1
        ),
    }
    state = GridPositionState(region=10, agents=agents)
    actor = GridMovementActor(position_state=state, agents=agents)

    for agent in agents.values():
        assert 'move' in agent.action_space

    state.reset()

    np.testing.assert_array_equal(
        actor.process_action(agents['agent0'], {'move': np.array([-1,  1])}), np.array([-1,  1])
    )
    np.testing.assert_array_equal(
        actor.process_action(agents['agent1'], {'move': np.array([0,  1])}), np.array([0,  1])
    )
    np.testing.assert_array_equal(
        actor.process_action(agents['agent2'], {'move': np.array([0, -1])}), np.array([0, -1])
    )
    np.testing.assert_array_equal(
        actor.process_action(agents['agent3'], {'move': np.array([1,  0])}), np.array([1,  0])
    )
    np.testing.assert_array_equal(agents['agent0'].position, np.array([5, 5]))
    np.testing.assert_array_equal(agents['agent1'].position, np.array([3, 4]))
    np.testing.assert_array_equal(agents['agent2'].position, np.array([0, 0]))
    np.testing.assert_array_equal(agents['agent3'].position, np.array([9, 4]))

    np.testing.assert_array_equal(
        actor.process_action(agents['agent0'], {'move': np.array([2, -2])}), np.array([2, -2])
    )
    np.testing.assert_array_equal(
        actor.process_action(agents['agent1'], {'move': np.array([-3,  0])}), np.array([-3,  0])
    )
    np.testing.assert_array_equal(
        actor.process_action(agents['agent2'], {'move': np.array([-1,  0])}), np.array([0,  0])
    )
    np.testing.assert_array_equal(
        actor.process_action(agents['agent3'], {'move': np.array([1,  1])}), np.array([0,  0])
    )
    np.testing.assert_array_equal(agents['agent0'].position, np.array([7, 3]))
    np.testing.assert_array_equal(agents['agent1'].position, np.array([0, 4]))
    np.testing.assert_array_equal(agents['agent2'].position, np.array([0, 0]))
    np.testing.assert_array_equal(agents['agent3'].position, np.array([9, 4]))


class SpeedAngleMovementTestAgent(SpeedAngleAgent, SpeedAngleActingAgent): pass


def test_speed_angle_movement_component():
    agents = {
        'agent0': SpeedAngleMovementTestAgent(
            id='agent0', initial_position=np.array([6.2,  3.3  ]), initial_speed=1.0,
            min_speed=0.0, max_speed=1.0, max_acceleration=0.35, initial_banking_angle=-30,
            max_banking_angle=45, max_banking_angle_change=30, initial_ground_angle=300
        ),
        'agent1': SpeedAngleMovementTestAgent(
            id='agent1', initial_position=np.array([2.1,  3.15 ]), initial_speed=0.5,
            min_speed=0.0, max_speed=1.0, max_acceleration=0.35, initial_banking_angle=10,
            max_banking_angle=45, max_banking_angle_change=30, initial_ground_angle=100
        ),
        'agent2': SpeedAngleMovementTestAgent(
            id='agent2', initial_position=np.array([0.5,  1.313]), initial_speed=0.24,
            min_speed=0.0, max_speed=1.0, max_acceleration=0.35, initial_banking_angle=0.0,
            max_banking_angle=45, max_banking_angle_change=30, initial_ground_angle=45
        ),
        'agent3': SpeedAngleMovementTestAgent(
            id='agent3', initial_position=np.array([8.24, 4.4  ]), initial_speed=0.6,
            min_speed=0.0, max_speed=1.0, max_acceleration=0.35, initial_banking_angle=24,
            max_banking_angle=45, max_banking_angle_change=30, initial_ground_angle=180
        ),
    }
    position_state = ContinuousPositionState(region=10, agents=agents)
    speed_angle_state = SpeedAngleState(agents=agents)
    actor = SpeedAngleMovementActor(
        position_state=position_state, speed_angle_state=speed_angle_state, agents=agents
    )

    for agent in agents.values():
        assert 'accelerate' in agent.action_space
        assert 'bank' in agent.action_space

    position_state.reset()
    speed_angle_state.reset()
    for agent in agents.values():
        np.testing.assert_array_equal(agent.position, agent.initial_position)
        assert agent.speed == agent.initial_speed
        assert agent.banking_angle == agent.initial_banking_angle
        assert agent.ground_angle == agent.initial_ground_angle

    assert np.allclose(
        actor.process_move(agents['agent0'], np.array([0.0]), np.array([0.0])), np.array([0., -1.])
    )
    assert np.allclose(agents['agent0'].position, np.array([6.2, 2.3]))
    assert np.allclose(
        actor.process_move(agents['agent1'], np.array([0.0]), np.array([0.0])),
        np.array([-0.17101007, 0.46984631])
    )
    assert np.allclose(agents['agent1'].position, np.array([1.92898993, 3.61984631]))
    assert np.allclose(
        actor.process_move(agents['agent2'], np.array([0.0]), np.array([0.0])),
        np.array([0.16970563, 0.16970563])
    )
    assert np.allclose(agents['agent2'].position, np.array([0.66970563, 1.48270563]))
    assert np.allclose(
        actor.process_move(agents['agent3'], np.array([0.0]), np.array([0.0])),
        np.array([-0.54812727, -0.24404199])
    )
    assert np.allclose(agents['agent3'].position, np.array([7.69187273, 4.15595801]))

    assert np.allclose(
        actor.process_move(agents['agent0'], np.array([-0.35]), np.array([30])),
        np.array([0, -0.65])
    )
    assert np.allclose(agents['agent0'].position, np.array([6.2,  1.65]))
    assert np.allclose(
        actor.process_move(agents['agent1'], np.array([-0.1]), np.array([-30])),
        np.array([0, 0.4])
    )
    assert np.allclose(agents['agent1'].position, np.array([1.92898993, 4.01984631]))
    assert np.allclose(
        actor.process_move(agents['agent2'], np.array([-0.24]), np.array([30])),
        np.array([0, 0.0])
    )
    assert np.allclose(agents['agent2'].position, np.array([0.66970563, 1.48270563]))
    assert np.allclose(
        actor.process_move(agents['agent3'],  np.array([0.0]), np.array([-24])),
        np.array([-0.54812727, -0.24404199])
    )
    assert np.allclose(agents['agent3'].position, np.array([7.14374545, 3.91191603]))


class ParticleAgent(VelocityAgent, AcceleratingAgent): pass


def test_acceleration_movement_component():
    agents = {
        'agent0': ParticleAgent(
            id='agent0', initial_position=np.array([2.3, 4.5]), initial_velocity=np.array([1, 0]),
            max_speed=1.0, max_acceleration=0.5
        ),
        'agent1': ParticleAgent(
            id='agent1', initial_position=np.array([8.5, 1.0]), initial_velocity=np.array([0, 0]),
            max_speed=1.0, max_acceleration=0.5
        ),
        'agent2': ParticleAgent(
            id='agent2', initial_position=np.array([5.0, 5.0]),
            initial_velocity=np.array([-1, -1]), max_speed=2.0, max_acceleration=0.5
        ),
    }

    position_state = ContinuousPositionState(region=10, agents=agents)
    velocity_state = VelocityState(agents=agents, friction=0.1)
    actor = AccelerationMovementActor(
        position_state=position_state, velocity_state=velocity_state, agents=agents
    )

    for agent in agents.values():
        assert 'accelerate' in agent.action_space

    position_state.reset()
    velocity_state.reset()
    for agent in agents.values():
        np.testing.assert_array_equal(agent.position, agent.initial_position)
        np.testing.assert_array_equal(agent.velocity, agent.initial_velocity)

    assert np.allclose(
        actor.process_action(agents['agent0'], {'accelerate': np.array([-1, 0])}),
        np.array([0., 0.])
    )
    assert np.allclose(agents['agent0'].position, np.array([2.3, 4.5]))
    assert np.allclose(
        actor.process_action(agents['agent1'], {'accelerate': np.array([-1, 1])}),
        np.array([-0.70710678,  0.70710678])
    )
    assert np.allclose(agents['agent1'].position, np.array([7.79289322, 1.70710678]))
    assert np.allclose(
        actor.process_action(agents['agent2'], {'accelerate': np.array([0, 0])}),
        np.array([-1, -1])
    )
    assert np.allclose(agents['agent2'].position, np.array([4, 4]))

    velocity_state.apply_friction(agents['agent0'])
    assert np.allclose(agents['agent0'].velocity, np.array([0, 0]))
    velocity_state.apply_friction(agents['agent1'])
    assert np.allclose(agents['agent1'].velocity, np.array([-0.6363961, 0.6363961]))



import numpy as np

from abmarl.sim.predator_prey import PredatorPreySimulation, Predator, Prey
from abmarl.sim.wrappers import CommunicationHandshakeWrapper


def test_communication():
    np.random.seed(24)
    agents = [
        Predator(id='predator0', view=1, attack=1),
        Predator(id='predator1', view=8, attack=0),
        Prey(id='prey1', view=4),
        Prey(id='prey2', view=5)
    ]
    sim = PredatorPreySimulation.build(
        {'agents': agents, 'observation_mode': PredatorPreySimulation.ObservationMode.DISTANCE}
    )
    sim = CommunicationHandshakeWrapper(sim)


    sim.reset()
    sim.sim.agents['prey1'].position = np.array([1, 1])
    sim.sim.agents['prey2'].position = np.array([1, 4])
    sim.sim.agents['predator0'].position = np.array([2, 3])
    sim.sim.agents['predator1'].position = np.array([0, 7])

    np.testing.assert_array_equal(
        sim.get_obs('predator0')['obs']['predator1'], np.array([0, 0, 0])
    )
    np.testing.assert_array_equal(
        sim.get_obs('predator0')['obs']['prey1'], np.array([0, 0, 0])
    )
    np.testing.assert_array_equal(
        sim.get_obs('predator0')['obs']['prey2'], np.array([-1, 1, 1])
    )
    assert sim.get_obs('predator0')['message_buffer'] == \
        {'predator1': False, 'prey1': False, 'prey2': False}

    np.testing.assert_array_equal(
        sim.get_obs('predator1')['obs']['predator0'], np.array([2, -4, 2])
    )
    np.testing.assert_array_equal(
        sim.get_obs('predator1')['obs']['prey1'], np.array([1, -6, 1])
    )
    np.testing.assert_array_equal(
        sim.get_obs('predator1')['obs']['prey2'], np.array([1, -3, 1])
    )
    assert sim.get_obs('predator1')['message_buffer'] == \
        {'predator0': False, 'prey1': False, 'prey2': False}

    np.testing.assert_array_equal(
        sim.get_obs('prey1')['obs']['predator0'], np.array([1, 2, 2])
    )
    np.testing.assert_array_equal(
        sim.get_obs('prey1')['obs']['predator1'], np.array([0, 0, 0])
    )
    np.testing.assert_array_equal(
        sim.get_obs('prey1')['obs']['prey2'], np.array([0, 3, 1])
    )
    assert sim.get_obs('prey1')['message_buffer'] == \
        {'predator0': False, 'predator1': False, 'prey2': False}

    np.testing.assert_array_equal(
        sim.get_obs('prey2')['obs']['predator0'], np.array([1, -1, 2])
    )
    np.testing.assert_array_equal(
        sim.get_obs('prey2')['obs']['predator1'], np.array([-1, 3, 2])
    )
    np.testing.assert_array_equal(
        sim.get_obs('prey2')['obs']['prey1'], np.array([0, -3, 1])
    )
    assert sim.get_obs('prey2')['message_buffer'] == \
        {'predator0': False, 'predator1': False, 'prey1': False}


    action1 = {
        'predator0': {
            'action': {'move': np.zeros(2), 'attack': 1},
            'send': {'predator1': False, 'prey1': False, 'prey2': False},
            'receive': {'predator1': False, 'prey1': False, 'prey2': True},
        },
        'predator1': {
            'action': {'move': np.zeros(2), 'attack': 0},
            'send': {'predator0': True, 'prey1': False, 'prey2': False},
            'receive': {'predator0': True, 'prey1': True, 'prey2': True},
        },
        'prey1': {
            'action': np.array([-1, 0]),
            'send': {'predator0': False, 'predator1': False, 'prey2': False},
            'receive': {'predator0': False, 'predator1': False, 'prey2': False},
        },
        'prey2': {
            'action': np.array([0, 1]),
            'send': {'predator0': False, 'predator1': False, 'prey1': True},
            'receive': {'predator0': True, 'predator1': True, 'prey1': True},
        }
    }
    sim.step(action1)

    np.testing.assert_array_equal(
        sim.get_obs('predator0')['obs']['predator1'], np.array([0, 0, 0])
    )
    np.testing.assert_array_equal(
        sim.get_obs('predator0')['obs']['prey1'], np.array([0, 0, 0])
    )
    np.testing.assert_array_equal(
        sim.get_obs('predator0')['obs']['prey2'], np.array([0, 0, 0])
    )
    assert sim.get_obs('predator0')['message_buffer'] == \
        {'predator1': True, 'prey1': False, 'prey2': False}
    assert sim.get_reward('predator0') == 100
    assert not sim.get_done('predator0')

    np.testing.assert_array_equal(
        sim.get_obs('predator1')['obs']['predator0'], np.array([2, -4, 2])
    )
    np.testing.assert_array_equal(
        sim.get_obs('predator1')['obs']['prey1'], np.array([0, -6, 1])
    )
    np.testing.assert_array_equal(
        sim.get_obs('predator1')['obs']['prey2'], np.array([0, 0, 0])
    )
    assert sim.get_obs('predator1')['message_buffer'] == \
        {'predator0': False, 'prey1': False, 'prey2': False}
    assert sim.get_reward('predator1') == 0
    assert not sim.get_done('predator1')

    np.testing.assert_array_equal(
        sim.get_obs('prey1')['obs']['predator0'], np.array([2, 2, 2])
    )
    np.testing.assert_array_equal(
        sim.get_obs('prey1')['obs']['predator1'], np.array([0, 0, 0])
    )
    np.testing.assert_array_equal(
        sim.get_obs('prey1')['obs']['prey2'], np.array([0, 0, 0])
    )
    assert sim.get_obs('prey1')['message_buffer'] == \
        {'predator0': False, 'predator1': False, 'prey2': True}
    assert sim.get_reward('prey1') == -1
    assert not sim.get_done('prey1')

    np.testing.assert_array_equal(
        sim.get_obs('prey2')['obs']['predator0'], np.array([1, -1, 2])
    )
    np.testing.assert_array_equal(
        sim.get_obs('prey2')['obs']['predator1'], np.array([-1, 3, 2])
    )
    np.testing.assert_array_equal(
        sim.get_obs('prey2')['obs']['prey1'], np.array([-1, -3, 1])
    )
    assert sim.get_obs('prey2')['message_buffer'] == \
        {'predator0': False, 'predator1': False, 'prey1': False}
    assert sim.get_reward('prey2') == -100
    assert sim.get_done('prey2')

    assert not sim.get_all_done()


    action2 = {
        'predator0': {
            'action': {'move': np.array([-1, 0]), 'attack': 0},
            'send': {'predator1': False, 'prey1': False, 'prey2': False},
            'receive': {'predator1': False, 'prey1': False, 'prey2': False}
        },
        'predator1': {
            'action': {'move': np.array([1, -1]), 'attack': 0},
            'send': {'predator0': True, 'prey1': False, 'prey2': False},
            'receive': {'predator0': True, 'prey1': True, 'prey2': False}
        },
        'prey1': {
            'action': np.array([1, -1]),
            'send': {'predator0': False, 'predator1': False, 'prey2': True},
            'receive': {'predator0': False, 'predator1': False, 'prey2': True}
        },
    }
    sim.step(action2)

    np.testing.assert_array_equal(
        sim.get_obs('predator0')['obs']['predator1'], np.array([0, 0, 0])
    )
    np.testing.assert_array_equal(
        sim.get_obs('predator0')['obs']['prey1'], np.array([0, 0, 0])
    )
    np.testing.assert_array_equal(
        sim.get_obs('predator0')['obs']['prey2'], np.array([0, 0, 0])
    )
    assert sim.get_obs('predator0')['message_buffer'] == \
        {'predator1': True, 'prey1': False, 'prey2': False}
    assert sim.get_reward('predator0') == -1
    assert not sim.get_done('predator0')

    np.testing.assert_array_equal(
        sim.get_obs('predator1')['obs']['predator0'], np.array([0, -3, 2])
    )
    np.testing.assert_array_equal(
        sim.get_obs('predator1')['obs']['prey1'], np.array([0, -6, 1])
    )
    np.testing.assert_array_equal(
        sim.get_obs('predator1')['obs']['prey2'], np.array([0, 0, 0])
    )
    assert sim.get_obs('predator1')['message_buffer'] == \
        {'predator0': False, 'prey1': False, 'prey2': False}
    assert sim.get_reward('predator1') == -1
    assert not sim.get_done('predator1')

    np.testing.assert_array_equal(
        sim.get_obs('prey1')['obs']['predator0'], np.array([0, 3, 2])
    )
    np.testing.assert_array_equal(
        sim.get_obs('prey1')['obs']['predator1'], np.array([0, 0, 0])
    )
    np.testing.assert_array_equal(
        sim.get_obs('prey1')['obs']['prey2'], np.array([0, 0, 0])
    )
    assert sim.get_obs('prey1')['message_buffer'] == \
        {'predator0': False, 'predator1': False, 'prey2': False}
    assert sim.get_reward('prey1') == -1
    assert not sim.get_done('prey1')


    action3 = {
        'predator0': {
            'action': {'move': np.array([0, -1]), 'attack': 0},
            'send': {'predator1': False, 'prey1': False, 'prey2': False},
            'receive': {'predator1': False, 'prey1': False, 'prey2': False},
        },
        'predator1': {
            'action': {'move': np.array([1, -1]), 'attack': 0},
            'send': {'predator0': True, 'prey1': False, 'prey2': False},
            'receieve': {'predator0': True, 'prey1': True, 'prey2': False}
        },
        'prey1': {
            'action': np.array([1, 0]),
            'send': {'predator0': False, 'predator1': False, 'prey2': True},
            'receive': {'predator0': False, 'predator1': False, 'prey2': True}
        }
    }
    sim.step(action3)

    np.testing.assert_array_equal(
        sim.get_obs('predator0')['obs']['predator1'], np.array([0, 0, 0])
    )
    np.testing.assert_array_equal(
        sim.get_obs('predator0')['obs']['prey1'], np.array([0, 0, 0])
    )
    np.testing.assert_array_equal(
        sim.get_obs('predator0')['obs']['prey2'], np.array([0, 0, 0])
    )
    assert sim.get_obs('predator0')['message_buffer'] == \
        {'predator1': True, 'prey1': False, 'prey2': False}
    assert sim.get_reward('predator0') == -1
    assert not sim.get_done('predator0')

    np.testing.assert_array_equal(
        sim.get_obs('predator1')['obs']['predator0'], np.array([-1, -3, 2])
    )
    np.testing.assert_array_equal(
        sim.get_obs('predator1')['obs']['prey1'], np.array([0, -5, 1])
    )
    np.testing.assert_array_equal(
        sim.get_obs('predator1')['obs']['prey2'], np.array([0, 0, 0])
    )
    assert sim.get_obs('predator1')['message_buffer'] == \
        {'predator0': False, 'prey1': False, 'prey2': False}
    assert sim.get_reward('predator1') == -1
    assert not sim.get_done('predator1')

    np.testing.assert_array_equal(
        sim.get_obs('prey1')['obs']['predator0'], np.array([-1, 2, 2])
    )
    np.testing.assert_array_equal(
        sim.get_obs('prey1')['obs']['predator1'], np.array([0, 0, 0])
    )
    np.testing.assert_array_equal(
        sim.get_obs('prey1')['obs']['prey2'], np.array([0, 0, 0])
    )
    assert sim.get_obs('prey1')['message_buffer'] == \
        {'predator0': False, 'predator1': False, 'prey2': False}
    assert sim.get_reward('prey1') == -1
    assert not sim.get_done('prey1')


    action4 = {
        'predator0': {
            'action': {'move': np.array([0, 0]), 'attack': 1},
            'send': {'predator1': False, 'prey1': False, 'prey2': False},
            'receive': {'predator1': False, 'prey1': True, 'prey2': False},
        },
        'predator1': {
            'action': {'move': np.array([1, -1]), 'attack': 0},
            'send': {'predator0': True, 'prey1': False, 'prey2': False},
            'receive': {'predator0': True, 'prey1': True, 'prey2': False}
        },
        'prey1': {
            'action': np.array([1, 0]),
            'send': {'predator0': False, 'predator1': False, 'prey2': False},
            'receive': {'predator0': True, 'predator1': True, 'prey2': True}
        },
    }
    sim.step(action4)

    np.testing.assert_array_equal(
        sim.get_obs('predator0')['obs']['predator1'], np.array([0, 0, 0])
    )
    np.testing.assert_array_equal(
        sim.get_obs('predator0')['obs']['prey1'], np.array([0, 0, 0])
    )
    np.testing.assert_array_equal(
        sim.get_obs('predator0')['obs']['prey2'], np.array([0, 0, 0])
    )
    assert sim.get_obs('predator0')['message_buffer'] == \
        {'predator1': True, 'prey1': False, 'prey2': False}
    assert sim.get_reward('predator0') == -10
    assert not sim.get_done('predator0')

    np.testing.assert_array_equal(
        sim.get_obs('predator1')['obs']['predator0'], np.array([-2, -2, 2])
    )
    np.testing.assert_array_equal(
        sim.get_obs('predator1')['obs']['prey1'], np.array([0, -4, 1])
    )
    np.testing.assert_array_equal(
        sim.get_obs('predator1')['obs']['prey2'], np.array([0, 0, 0])
    )
    assert sim.get_obs('predator1')['message_buffer'] == \
        {'predator0': False, 'prey1': False, 'prey2': False}
    assert sim.get_reward('predator1') == -1
    assert not sim.get_done('predator1')

    np.testing.assert_array_equal(
        sim.get_obs('prey1')['obs']['predator0'], np.array([-2, 2, 2])
    )
    np.testing.assert_array_equal(
        sim.get_obs('prey1')['obs']['predator1'], np.array([0, 4, 2])
    )
    np.testing.assert_array_equal(
        sim.get_obs('prey1')['obs']['prey2'], np.array([0, 0, 0])
    )
    assert sim.get_obs('prey1')['message_buffer'] == \
        {'predator0': False, 'predator1': False, 'prey2': False}
    assert sim.get_reward('prey1') == -1
    assert not sim.get_done('prey1')


    action5 = {
        'predator0': {
            'action': {'move': np.zeros(2), 'attack': 0},
            'send': {'predator1': True, 'prey1': False, 'prey2': False},
            'receive': {'predator1': True, 'prey1': False, 'prey2': False},
        },
        'predator1': {
            'action': {'move': np.array([1, 0]), 'attack': 0},
            'send': {'predator0': True, 'prey1': False, 'prey2': False},
            'receive': {'predator0': True, 'prey1': True, 'prey2': False},
        },
        'prey1': {
            'action': np.array([0, -1]),
            'send': {'predator0': False, 'predator1': False, 'prey2': False},
            'receive': {'predator0': False, 'predator1': False, 'prey2': False},
        },
    }
    sim.step(action5)

    np.testing.assert_array_equal(
        sim.get_obs('predator0')['obs']['predator1'], np.array([3, 2, 2])
    )
    np.testing.assert_array_equal(
        sim.get_obs('predator0')['obs']['prey1'], np.array([2, -2, 1])
    )
    np.testing.assert_array_equal(
        sim.get_obs('predator0')['obs']['prey2'], np.array([0, 0, 0])
    )
    assert sim.get_obs('predator0')['message_buffer'] == \
        {'predator1': True, 'prey1': False, 'prey2': False}
    assert sim.get_reward('predator0') == 0
    assert not sim.get_done('predator0')

    np.testing.assert_array_equal(
        sim.get_obs('predator1')['obs']['predator0'], np.array([-3, -2, 2])
    )
    np.testing.assert_array_equal(
        sim.get_obs('predator1')['obs']['prey1'], np.array([-1, -4, 1])
    )
    np.testing.assert_array_equal(
        sim.get_obs('predator1')['obs']['prey2'], np.array([0, 0, 0])
    )
    assert sim.get_obs('predator1')['message_buffer'] == \
        {'predator0': True, 'prey1': False, 'prey2': False}
    assert sim.get_reward('predator1') == -1
    assert not sim.get_done('predator1')

    np.testing.assert_array_equal(
        sim.get_obs('prey1')['obs']['predator0'], np.array([-2, 2, 2])
    )
    np.testing.assert_array_equal(
        sim.get_obs('prey1')['obs']['predator1'], np.array([1, 4, 2])
    )
    np.testing.assert_array_equal(
        sim.get_obs('prey1')['obs']['prey2'], np.array([0, 0, 0])
    )
    assert sim.get_obs('prey1')['message_buffer'] == \
        {'predator0': False, 'predator1': False, 'prey2': False}
    assert sim.get_reward('prey1') == -10
    assert not sim.get_done('prey1')

    action6 = {
        'predator0': {
            'action': {'move': np.array([1, 0]), 'attack': 0},
            'send': {'predator1': True, 'prey1': False, 'prey2': False},
            'receive': {'predator1': True, 'prey1': False, 'prey2': False}
        },
        'predator1': {
            'action': {'move': np.array([1, 0]), 'attack': 0},
            'send': {'predator0': True, 'prey1': False, 'prey2': False},
            'receive': {'predator0': True, 'prey1': True, 'prey2': False}
        },
        'prey1': {
            'action': np.array([1, 1]),
            'send': {'predator0': False, 'predator1': False, 'prey2': False},
            'receive': {'predator0': False, 'predator1': False, 'prey2': False},
        }
    }
    sim.step(action6)

    np.testing.assert_array_equal(
        sim.get_obs('predator0')['obs']['predator1'], np.array([3, 2, 2])
    )
    np.testing.assert_array_equal(
        sim.get_obs('predator0')['obs']['prey1'], np.array([2, -1, 1])
    )
    np.testing.assert_array_equal(
        sim.get_obs('predator0')['obs']['prey2'], np.array([0, 0, 0])
    )
    assert sim.get_obs('predator0')['message_buffer'] == \
        {'predator1': True, 'prey1': False, 'prey2': False}
    assert sim.get_reward('predator0') == -1
    assert not sim.get_done('predator0')

    np.testing.assert_array_equal(
        sim.get_obs('predator1')['obs']['predator0'], np.array([-3, -2, 2])
    )
    np.testing.assert_array_equal(
        sim.get_obs('predator1')['obs']['prey1'], np.array([-1, -3, 1])
    )
    np.testing.assert_array_equal(
        sim.get_obs('predator1')['obs']['prey2'], np.array([0, 0, 0])
    )
    assert sim.get_obs('predator1')['message_buffer'] == \
        {'predator0': True, 'prey1': False, 'prey2': False}
    assert sim.get_reward('predator1') == -1
    assert not sim.get_done('predator1')

    np.testing.assert_array_equal(
        sim.get_obs('prey1')['obs']['predator0'], np.array([-2, 1, 2])
    )
    np.testing.assert_array_equal(
        sim.get_obs('prey1')['obs']['predator1'], np.array([1, 3, 2])
    )
    np.testing.assert_array_equal(
        sim.get_obs('prey1')['obs']['prey2'], np.array([0, 0, 0])
    )
    assert sim.get_obs('prey1')['message_buffer'] == \
        {'predator0': False, 'predator1': False, 'prey2': False}
    assert sim.get_reward('prey1') == -1
    assert not sim.get_done('prey1')


    action7 = {
        'predator0': {
            'action': {'move': np.array([1, 0]), 'attack': 0},
            'send': {'predator1': True, 'prey1': False, 'prey2': False},
            'receive': {'predator1': True, 'prey1': False, 'prey2': False}
        },
        'predator1': {
            'action': {'move': np.array([1, 1]), 'attack': 0},
            'send': {'predator0': True, 'prey1': False, 'prey2': False},
            'receive': {'predator0': True, 'prey1': True, 'prey2': False}
        },
        'prey1': {
            'action': np.array([1, 1]),
            'send': {'predator0': False, 'predator1': False, 'prey2': False},
            'receive': {'predator0': False, 'predator1': False, 'prey2': False},
        }
    }
    sim.step(action7)

    np.testing.assert_array_equal(
        sim.get_obs('predator0')['obs']['predator1'], np.array([3, 3, 2])
    )
    np.testing.assert_array_equal(
        sim.get_obs('predator0')['obs']['prey1'], np.array([2, 0, 1])
    )
    np.testing.assert_array_equal(
        sim.get_obs('predator0')['obs']['prey2'], np.array([0, 0, 0])
    )
    assert sim.get_obs('predator0')['message_buffer'] == \
        {'predator1': True, 'prey1': False, 'prey2': False}
    assert sim.get_reward('predator0') == -1
    assert not sim.get_done('predator0')

    np.testing.assert_array_equal(
        sim.get_obs('predator1')['obs']['predator0'], np.array([-3, -3, 2])
    )
    np.testing.assert_array_equal(
        sim.get_obs('predator1')['obs']['prey1'], np.array([-1, -3, 1])
    )
    np.testing.assert_array_equal(
        sim.get_obs('predator1')['obs']['prey2'], np.array([0, 0, 0])
    )
    assert sim.get_obs('predator1')['message_buffer'] == \
        {'predator0': True, 'prey1': False, 'prey2': False}
    assert sim.get_reward('predator1') == -1
    assert not sim.get_done('predator1')

    np.testing.assert_array_equal(
        sim.get_obs('prey1')['obs']['predator0'], np.array([-2, 0, 2])
    )
    np.testing.assert_array_equal(
        sim.get_obs('prey1')['obs']['predator1'], np.array([1, 3, 2])
    )
    np.testing.assert_array_equal(
        sim.get_obs('prey1')['obs']['prey2'], np.array([0, 0, 0])
    )
    assert sim.get_obs('prey1')['message_buffer'] == \
        {'predator0': False, 'predator1': False, 'prey2': False}
    assert sim.get_reward('prey1') == -1
    assert not sim.get_done('prey1')


    action8 = {
        'predator0': {
            'action': {'move': np.array([1, 0]), 'attack': 0},
            'send': {'predator1': True, 'prey1': False, 'prey2': False},
            'receive': {'predator1': True, 'prey1': False, 'prey2': False}
        },
        'predator1': {
            'action': {'move': np.array([-1, -1]), 'attack': 0},
            'send': {'predator0': True, 'prey1': False, 'prey2': False},
            'receive': {'predator0': True, 'prey1': True, 'prey2': False}
        },
        'prey1': {
            'action': np.array([0, 1]),
            'send': {'predator0': False, 'predator1': False, 'prey2': False},
            'receive': {'predator0': False, 'predator1': False, 'prey2': False},
        }
    }
    sim.step(action8)

    np.testing.assert_array_equal(
        sim.get_obs('predator0')['obs']['predator1'], np.array([1, 2, 2])
    )
    np.testing.assert_array_equal(
        sim.get_obs('predator0')['obs']['prey1'], np.array([1, 1, 1])
    )
    np.testing.assert_array_equal(
        sim.get_obs('predator0')['obs']['prey2'], np.array([0, 0, 0])
    )
    assert sim.get_obs('predator0')['message_buffer'] == \
        {'predator1': True, 'prey1': False, 'prey2': False}
    assert sim.get_reward('predator0') == -1
    assert not sim.get_done('predator0')

    np.testing.assert_array_equal(
        sim.get_obs('predator1')['obs']['predator0'], np.array([-1, -2, 2])
    )
    np.testing.assert_array_equal(
        sim.get_obs('predator1')['obs']['prey1'], np.array([0, -1, 1])
    )
    np.testing.assert_array_equal(
        sim.get_obs('predator1')['obs']['prey2'], np.array([0, 0, 0])
    )
    assert sim.get_obs('predator1')['message_buffer'] == \
        {'predator0': True, 'prey1': False, 'prey2': False}
    assert sim.get_reward('predator1') == -1
    assert not sim.get_done('predator1')

    np.testing.assert_array_equal(
        sim.get_obs('prey1')['obs']['predator0'], np.array([-1, -1, 2])
    )
    np.testing.assert_array_equal(
        sim.get_obs('prey1')['obs']['predator1'], np.array([0, 1, 2])
    )
    np.testing.assert_array_equal(
        sim.get_obs('prey1')['obs']['prey2'], np.array([0, 0, 0])
    )
    assert sim.get_obs('prey1')['message_buffer'] == \
        {'predator0': False, 'predator1': False, 'prey2': False}
    assert sim.get_reward('prey1') == -1
    assert not sim.get_done('prey1')


    action9 = {
        'predator0': {
            'action': {'move': np.array([0, 0]), 'attack': 1},
            'send': {'predator1': True, 'prey1': False, 'prey2': False},
            'receive': {'predator1': True, 'prey1': False, 'prey2': False}
        },
        'predator1': {
            'action': {'move': np.array([0, 0]), 'attack': 0},
            'send': {'predator0': True, 'prey1': False, 'prey2': False},
            'receive': {'predator0': True, 'prey1': True, 'prey2': False}
        },
        'prey1': {
            'action': np.array([-1, 1]),
            'send': {'predator0': False, 'predator1': False, 'prey2': False},
            'receive': {'predator0': False, 'predator1': False, 'prey2': False},
        }
    }
    sim.step(action9)

    np.testing.assert_array_equal(
        sim.get_obs('predator0')['obs']['predator1'], np.array([1, 2, 2])
    )
    np.testing.assert_array_equal(
        sim.get_obs('predator0')['obs']['prey1'], np.array([0, 0, 0])
    )
    np.testing.assert_array_equal(
        sim.get_obs('predator0')['obs']['prey2'], np.array([0, 0, 0])
    )
    assert sim.get_obs('predator0')['message_buffer'] == \
        {'predator1': True, 'prey1': False, 'prey2': False}
    assert sim.get_reward('predator0') == 100
    assert not sim.get_done('predator0')

    np.testing.assert_array_equal(
        sim.get_obs('predator1')['obs']['predator0'], np.array([-1, -2, 2])
    )
    np.testing.assert_array_equal(
        sim.get_obs('predator1')['obs']['prey1'], np.array([0, 0, 0])
    )
    np.testing.assert_array_equal(
        sim.get_obs('predator1')['obs']['prey2'], np.array([0, 0, 0])
    )
    assert sim.get_obs('predator1')['message_buffer'] == \
        {'predator0': True, 'prey1': False, 'prey2': False}
    assert sim.get_reward('predator1') == 0
    assert not sim.get_done('predator1')

    np.testing.assert_array_equal(
        sim.get_obs('prey1')['obs']['predator0'], np.array([-1, -1, 2])
    )
    np.testing.assert_array_equal(
        sim.get_obs('prey1')['obs']['predator1'], np.array([0, 1, 2])
    )
    np.testing.assert_array_equal(
        sim.get_obs('prey1')['obs']['prey2'], np.array([0, 0, 0])
    )
    assert sim.get_obs('prey1')['message_buffer'] == \
        {'predator0': False, 'predator1': False, 'prey2': False}
    assert sim.get_reward('prey1') == -100
    assert sim.get_done('prey1')

    assert sim.get_all_done()



import numpy as np

from abmarl.sim.components.agent import ComponentAgent as Agent
from abmarl.sim.components.state import LifeState, ContinuousPositionState
from abmarl.sim.components.done import DeadDone, TeamDeadDone, TooCloseDone


def test_dead_done_condition():
    agents = {
        'agent0': Agent(id='agent0'),
        'agent1': Agent(id='agent1'),
        'agent2': Agent(id='agent2'),
        'agent3': Agent(id='agent3'),
    }
    state = LifeState(agents=agents)
    done = DeadDone(agents=agents)
    state.reset()

    assert not done.get_done(agents['agent0'])
    assert not done.get_done(agents['agent1'])
    assert not done.get_done(agents['agent2'])
    assert not done.get_done(agents['agent3'])
    assert not done.get_all_done()

    agents['agent0'].is_alive = False
    agents['agent1'].is_alive = False
    assert done.get_done(agents['agent0'])
    assert done.get_done(agents['agent1'])
    assert not done.get_done(agents['agent2'])
    assert not done.get_done(agents['agent3'])
    assert not done.get_all_done()

    agents['agent2'].is_alive = False
    agents['agent3'].is_alive = False
    assert done.get_done(agents['agent0'])
    assert done.get_done(agents['agent1'])
    assert done.get_done(agents['agent2'])
    assert done.get_done(agents['agent3'])
    assert done.get_all_done()


def test_team_dead_done_condition():
    agents = {
        'agent0': Agent(id='agent0', team=1),
        'agent1': Agent(id='agent1', team=2),
        'agent2': Agent(id='agent2', team=1),
        'agent3': Agent(id='agent3', team=2),
        'agent4': Agent(id='agent4', team=1),
        'agent5': Agent(id='agent5', team=3),
    }
    state = LifeState(agents=agents)
    done = TeamDeadDone(agents=agents, number_of_teams=3)
    state.reset()

    assert not done.get_done(agents['agent0'])
    assert not done.get_done(agents['agent1'])
    assert not done.get_done(agents['agent2'])
    assert not done.get_done(agents['agent3'])
    assert not done.get_done(agents['agent4'])
    assert not done.get_done(agents['agent5'])
    assert not done.get_all_done()

    agents['agent5'].is_alive = False
    agents['agent4'].is_alive = False
    assert not done.get_done(agents['agent0'])
    assert not done.get_done(agents['agent1'])
    assert not done.get_done(agents['agent2'])
    assert not done.get_done(agents['agent3'])
    assert done.get_done(agents['agent4'])
    assert done.get_done(agents['agent5'])
    assert not done.get_all_done()

    agents['agent1'].is_alive = False
    agents['agent3'].is_alive = False
    assert not done.get_done(agents['agent0'])
    assert done.get_done(agents['agent1'])
    assert not done.get_done(agents['agent2'])
    assert done.get_done(agents['agent3'])
    assert done.get_done(agents['agent4'])
    assert done.get_done(agents['agent5'])
    assert done.get_all_done()


def test_too_close_done_with_continuous():
    agents = {
        'agent0': Agent(id='agent0', initial_position=np.array([0.1, 0.1])),
        'agent1': Agent(id='agent1', initial_position=np.array([0.24, 0.5])),
        'agent2': Agent(id='agent2', initial_position=np.array([0.3, 0.5])),
        'agent3': Agent(id='agent3', initial_position=np.array([3.76, 3.5])),
        'agent4': Agent(id='agent4', initial_position=np.array([3.75, 3.6])),
        'agent5': Agent(id='agent5', initial_position=np.array([2.5, 3.0])),
    }

    state = ContinuousPositionState(region=4, agents=agents)
    done = TooCloseDone(position=state, agents=agents, collision_distance=0.25)
    state.reset()

    assert done.get_done(agents['agent0'])
    assert done.get_done(agents['agent1'])
    assert done.get_done(agents['agent2'])
    assert done.get_done(agents['agent3'])
    assert done.get_done(agents['agent4'])
    assert not done.get_done(agents['agent5'])
    assert done.get_all_done()



import numpy as np

from abmarl.sim.components.agent import SpeedAngleAgent, VelocityAgent, BroadcastingAgent
from abmarl.sim.components.state import GridPositionState, LifeState, SpeedAngleState, \
    VelocityState, BroadcastState
from abmarl.sim.components.observer import HealthObserver, LifeObserver, PositionObserver, \
    RelativePositionObserver, SpeedObserver, AngleObserver, VelocityObserver, TeamObserver
from abmarl.sim.components.wrappers.observer_wrapper import \
    PositionRestrictedObservationWrapper, TeamBasedCommunicationWrapper
from abmarl.sim.components.actor import BroadcastActor

from abmarl.sim.components.agent import AgentObservingAgent, VelocityObservingAgent, \
    PositionObservingAgent, SpeedAngleObservingAgent, TeamObservingAgent, LifeObservingAgent, \
    HealthObservingAgent


class AllObservingAgent(
    AgentObservingAgent, VelocityObservingAgent, PositionObservingAgent, SpeedAngleObservingAgent,
    TeamObservingAgent, LifeObservingAgent, HealthObservingAgent
): pass


class NonViewAgent(
    VelocityObservingAgent, PositionObservingAgent, SpeedAngleObservingAgent, TeamObservingAgent,
    LifeObservingAgent, HealthObservingAgent
): pass


class AllAgent(           AllObservingAgent, SpeedAngleAgent, VelocityAgent): pass
class SpeedAnglelessAgent(AllObservingAgent,                  VelocityAgent): pass
class VelocitylessAgent(  AllObservingAgent, SpeedAngleAgent               ): pass


def test_position_restricted_observer_wrapper():
    agents = {
        'agent0': AllAgent(
            id='agent0', agent_view=2, initial_position=np.array([2, 2]), initial_health=0.67,
            team=1, max_speed=1, initial_speed=0.30, initial_banking_angle=7,
            initial_ground_angle=123, initial_velocity=np.array([-0.3, 0.8])
        ),
        'agent1': AllAgent(
            id='agent1', agent_view=1, initial_position=np.array([4, 4]), initial_health=0.54,
            team=2, max_speed=2, initial_speed=0.00, initial_banking_angle=0,
            initial_ground_angle=126, initial_velocity=np.array([-0.2, 0.7])
        ),
        'agent2': AllAgent(
            id='agent2', agent_view=1, initial_position=np.array([4, 3]), initial_health=0.36,
            team=2, max_speed=1, initial_speed=0.12, initial_banking_angle=30,
            initial_ground_angle=180, initial_velocity=np.array([-0.1, 0.6])
        ),
        'agent3': AllAgent(
            id='agent3', agent_view=1, initial_position=np.array([4, 2]), initial_health=0.24,
            team=2, max_speed=4, initial_speed=0.05, initial_banking_angle=13,
            initial_ground_angle=46,  initial_velocity=np.array([0.0,  0.5])
        ),
        'agent6': SpeedAnglelessAgent(
            id='agent6', agent_view=0, initial_position=np.array([1, 1]), initial_health=0.53,
            team=1, max_speed=1, initial_velocity=np.array([0.3, 0.2])
        ),
        'agent7': VelocitylessAgent(
            id='agent7', agent_view=5, initial_position=np.array([0, 4]), initial_health=0.50,
            team=2, max_speed=1, initial_speed=0.36, initial_banking_angle=24,
            initial_ground_angle=0
        ),
    }

    def linear_drop_off(distance, view):
        return 1. - 1. / (view+1) * distance

    np.random.seed(12)
    position_state = GridPositionState(agents=agents, region=5)
    life_state = LifeState(agents=agents)
    speed_state = SpeedAngleState(agents=agents)
    angle_state = SpeedAngleState(agents=agents)
    velocity_state = VelocityState(agents=agents)

    position_observer = PositionObserver(position_state=position_state, agents=agents)
    relative_position_observer = RelativePositionObserver(
        position_state=position_state, agents=agents
    )
    health_observer = HealthObserver(agents=agents)
    life_observer = LifeObserver(agents=agents)
    team_observer = TeamObserver(number_of_teams=3, agents=agents)
    speed_observer = SpeedObserver(agents=agents)
    angle_observer = AngleObserver(agents=agents)
    velocity_observer = VelocityObserver(agents=agents)

    position_state.reset()
    life_state.reset()
    speed_state.reset()
    angle_state.reset()
    velocity_state.reset()

    wrapped_observer = PositionRestrictedObservationWrapper(
        [
            position_observer,
            relative_position_observer,
            health_observer,
            life_observer,
            team_observer,
            speed_observer,
            angle_observer,
            velocity_observer
        ],
        obs_filter=linear_drop_off,
        agents=agents
    )

    obs = wrapped_observer.get_obs(agents['agent0'])
    assert obs['health'] == {
        'agent0': 0.67,
        'agent1': -1,
        'agent2': 0.36,
        'agent3': -1,
        'agent6': 0.53,
        'agent7': -1,
    }
    assert obs['life'] == {
        'agent0': True,
        'agent1': -1,
        'agent2': True,
        'agent3': -1,
        'agent6': 1,
        'agent7': -1,
    }
    assert obs['team'] == {
        'agent0': 1,
        'agent1': -1,
        'agent2': 2,
        'agent3': -1,
        'agent6': 1,
        'agent7': -1,
    }
    np.testing.assert_array_equal(obs['position']['agent0'], np.array([2, 2]))
    np.testing.assert_array_equal(obs['position']['agent1'], np.array([-1, -1]))
    np.testing.assert_array_equal(obs['position']['agent2'], np.array([4, 3]))
    np.testing.assert_array_equal(obs['position']['agent3'], np.array([-1, -1]))
    np.testing.assert_array_equal(obs['position']['agent6'], np.array([1, 1]))
    np.testing.assert_array_equal(obs['position']['agent7'], np.array([-1, -1]))
    np.testing.assert_array_equal(obs['relative_position']['agent0'], np.array([0, 0]))
    np.testing.assert_array_equal(obs['relative_position']['agent1'], np.array([-5, -5]))
    np.testing.assert_array_equal(obs['relative_position']['agent2'], np.array([2, 1]))
    np.testing.assert_array_equal(obs['relative_position']['agent3'], np.array([-5, -5]))
    np.testing.assert_array_equal(obs['relative_position']['agent6'], np.array([-1, -1]))
    np.testing.assert_array_equal(obs['relative_position']['agent7'], np.array([-5, -5]))
    assert obs['mask'] == {
        'agent0': 1,
        'agent1': 0,
        'agent2': 1,
        'agent3': 0,
        'agent6': 1,
        'agent7': 0,
    }
    assert obs['speed'] == {
        'agent0': 0.3,
        'agent1': -1,
        'agent2': 0.12,
        'agent3': -1,
        'agent6': -1,
        'agent7': -1,
    }
    assert obs['ground_angle'] == {
        'agent0': 123,
        'agent1': -1,
        'agent2': 180,
        'agent3': -1,
        'agent6': -1,
        'agent7': -1,
    }
    np.testing.assert_array_equal(obs['velocity']['agent0'], np.array([-0.3, 0.8]))
    np.testing.assert_array_equal(obs['velocity']['agent1'], np.array([0.0, 0.0]))
    np.testing.assert_array_equal(obs['velocity']['agent2'], np.array([-0.1, 0.6]))
    np.testing.assert_array_equal(obs['velocity']['agent3'], np.array([0.0, 0.0]))
    np.testing.assert_array_equal(obs['velocity']['agent6'], np.array([0.3, 0.2]))
    np.testing.assert_array_equal(obs['velocity']['agent7'], np.array([0.0, 0.0]))


    obs = wrapped_observer.get_obs(agents['agent1'])
    assert obs['health'] == {
        'agent0': -1,
        'agent1': 0.54,
        'agent2': -1,
        'agent3': -1,
        'agent6': -1,
        'agent7': -1,
    }
    assert obs['life'] == {
        'agent0': -1,
        'agent1': True,
        'agent2': -1,
        'agent3': -1,
        'agent6': -1,
        'agent7': -1,
    }
    assert obs['team'] == {
        'agent0': -1,
        'agent1': 2,
        'agent2': -1,
        'agent3': -1,
        'agent6': -1,
        'agent7': -1,
    }
    np.testing.assert_array_equal(obs['position']['agent0'], np.array([-1, -1]))
    np.testing.assert_array_equal(obs['position']['agent1'], np.array([4, 4]))
    np.testing.assert_array_equal(obs['position']['agent2'], np.array([-1, -1]))
    np.testing.assert_array_equal(obs['position']['agent3'], np.array([-1, -1]))
    np.testing.assert_array_equal(obs['position']['agent6'], np.array([-1, -1]))
    np.testing.assert_array_equal(obs['position']['agent7'], np.array([-1, -1]))
    np.testing.assert_array_equal(obs['relative_position']['agent0'], np.array([-5, -5]))
    np.testing.assert_array_equal(obs['relative_position']['agent1'], np.array([0, 0]))
    np.testing.assert_array_equal(obs['relative_position']['agent2'], np.array([-5, -5]))
    np.testing.assert_array_equal(obs['relative_position']['agent3'], np.array([-5, -5]))
    np.testing.assert_array_equal(obs['relative_position']['agent6'], np.array([-5, -5]))
    np.testing.assert_array_equal(obs['relative_position']['agent7'], np.array([-5, -5]))
    assert obs['mask'] == {
        'agent0': 0,
        'agent1': 1,
        'agent2': 0,
        'agent3': 0,
        'agent6': 0,
        'agent7': 0,
    }
    assert obs['speed'] == {
        'agent0': -1,
        'agent1': 0.0,
        'agent2': -1,
        'agent3': -1,
        'agent6': -1,
        'agent7': -1,
    }
    assert obs['ground_angle'] == {
        'agent0': -1,
        'agent1': 126,
        'agent2': -1,
        'agent3': -1,
        'agent6': -1,
        'agent7': -1,
    }
    np.testing.assert_array_equal(obs['velocity']['agent0'], np.array([0.0, 0.0]))
    np.testing.assert_array_equal(obs['velocity']['agent1'], np.array([-0.2, 0.7]))
    np.testing.assert_array_equal(obs['velocity']['agent2'], np.array([0.0, 0.0]))
    np.testing.assert_array_equal(obs['velocity']['agent3'], np.array([0.0, 0.0]))
    np.testing.assert_array_equal(obs['velocity']['agent6'], np.array([0.0, 0.0]))
    np.testing.assert_array_equal(obs['velocity']['agent7'], np.array([0.0, 0.0]))


class CommunicatingAgent(
    BroadcastingAgent, PositionObservingAgent, TeamObservingAgent, AgentObservingAgent
): pass


def test_broadcast_communication_observer_wrapper():
    agents = {
        'agent0': CommunicatingAgent(
            id='agent0', initial_position=np.array([1, 7]), team=1, broadcast_range=0,
            agent_view=0
        ),
        'agent1': CommunicatingAgent(
            id='agent1', initial_position=np.array([3, 3]), team=1, broadcast_range=4,
            agent_view=3
        ),
        'agent2': CommunicatingAgent(
            id='agent2', initial_position=np.array([5, 0]), team=2, broadcast_range=4,
            agent_view=2
        ),
        'agent3': CommunicatingAgent(
            id='agent3', initial_position=np.array([6, 9]), team=2, broadcast_range=4,
            agent_view=2
        ),
        'agent4': CommunicatingAgent(
            id='agent4', initial_position=np.array([4, 7]), team=2, broadcast_range=4,
            agent_view=3
        ),
    }

    position_state = GridPositionState(region=10, agents=agents)
    broadcast_state = BroadcastState(agents=agents)

    position_observer = PositionObserver(position_state=position_state, agents=agents)
    team_observer = TeamObserver(number_of_teams=2, agents=agents)
    partial_observer = PositionRestrictedObservationWrapper(
        [position_observer, team_observer], agents=agents
    )
    comms_observer = TeamBasedCommunicationWrapper([partial_observer], agents=agents)

    broadcast_actor = BroadcastActor(broadcast_state=broadcast_state, agents=agents)

    position_state.reset()
    broadcast_state.reset()

    action_dict = {
        'agent0': {'broadcast': 0},
        'agent1': {'broadcast': 1},
        'agent2': {'broadcast': 0},
        'agent3': {'broadcast': 0},
        'agent4': {'broadcast': 1},
    }
    for agent_id, action in action_dict.items():
        broadcast_actor.process_action(agents[agent_id], action)

    obs = partial_observer.get_obs(agents['agent0'])
    assert obs['mask'] == {
        'agent0': 1,
        'agent1': 0,
        'agent2': 0,
        'agent3': 0,
        'agent4': 0,
    }
    assert obs['team'] == {
        'agent0': 1,
        'agent1': -1,
        'agent2': -1,
        'agent3': -1,
        'agent4': -1,
    }
    np.testing.assert_array_equal(obs['position']['agent0'], np.array([1, 7]))
    np.testing.assert_array_equal(obs['position']['agent1'], np.array([-1, -1]))
    np.testing.assert_array_equal(obs['position']['agent2'], np.array([-1, -1]))
    np.testing.assert_array_equal(obs['position']['agent3'], np.array([-1, -1]))
    np.testing.assert_array_equal(obs['position']['agent4'], np.array([-1, -1]))

    obs = comms_observer.get_obs(agents['agent0'])
    assert obs['mask'] == {
        'agent0': 1,
        'agent1': 1,
        'agent2': 1,
        'agent3': 0,
        'agent4': 1,
    }
    assert obs['team'] == {
        'agent0': 1,
        'agent1': 1,
        'agent2': 2,
        'agent3': -1,
        'agent4': 2,
    }
    np.testing.assert_array_equal(obs['position']['agent0'], np.array([1, 7]))
    np.testing.assert_array_equal(obs['position']['agent1'], np.array([3, 3]))
    np.testing.assert_array_equal(obs['position']['agent2'], np.array([5, 0]))
    np.testing.assert_array_equal(obs['position']['agent3'], np.array([-1, -1]))
    np.testing.assert_array_equal(obs['position']['agent4'], np.array([4, 7]))


    action_dict = {
        'agent0': {'broadcast': 0},
        'agent1': {'broadcast': 0},
        'agent2': {'broadcast': 1},
        'agent3': {'broadcast': 1},
        'agent4': {'broadcast': 0},
    }
    for agent_id, action in action_dict.items():
        broadcast_actor.process_action(agents[agent_id], action)

    obs = comms_observer.get_obs(agents['agent0'])
    assert obs['mask'] == {
        'agent0': 1,
        'agent1': 0,
        'agent2': 0,
        'agent3': 0,
        'agent4': 0,
    }
    assert obs['team'] == {
        'agent0': 1,
        'agent1': -1,
        'agent2': -1,
        'agent3': -1,
        'agent4': -1,
    }
    np.testing.assert_array_equal(obs['position']['agent0'], np.array([1, 7]))
    np.testing.assert_array_equal(obs['position']['agent1'], np.array([-1, -1]))
    np.testing.assert_array_equal(obs['position']['agent2'], np.array([-1, -1]))
    np.testing.assert_array_equal(obs['position']['agent3'], np.array([-1, -1]))
    np.testing.assert_array_equal(obs['position']['agent4'], np.array([-1, -1]))



import numpy as np

from abmarl.sim.components.agent import AttackingAgent
from abmarl.sim.components.actor import AttackActor


def test_position_based_attack_actor():
    agents = {
        'agent0': AttackingAgent(
            id='agent0', attack_range=1, initial_position=np.array([1, 1]), attack_strength=0.6
        ),
        'agent1': AttackingAgent(
            id='agent1', attack_range=1, initial_position=np.array([0, 1]), attack_strength=0.6
        ),
        'agent2': AttackingAgent(
            id='agent2', attack_range=1, initial_position=np.array([4, 2]), attack_strength=0.6
        ),
        'agent3': AttackingAgent(
            id='agent3', attack_range=1, initial_position=np.array([4, 3]), attack_strength=0.6
        ),
        'agent4': AttackingAgent(
            id='agent4', attack_range=0, initial_position=np.array([3, 2]), attack_strength=0.6
        ),
        'agent5': AttackingAgent(
            id='agent5', attack_range=2, initial_position=np.array([4, 0]), attack_strength=0.6
        ),
    }

    assert agents['agent0'].attack_range == 1
    np.testing.assert_array_equal(agents['agent0'].initial_position, np.array([1, 1]))
    assert agents['agent1'].attack_range == 1
    np.testing.assert_array_equal(agents['agent1'].initial_position, np.array([0, 1]))
    assert agents['agent2'].attack_range == 1
    np.testing.assert_array_equal(agents['agent2'].initial_position, np.array([4, 2]))
    assert agents['agent3'].attack_range == 1
    np.testing.assert_array_equal(agents['agent3'].initial_position, np.array([4, 3]))
    assert agents['agent4'].attack_range == 0
    np.testing.assert_array_equal(agents['agent4'].initial_position, np.array([3, 2]))
    assert agents['agent5'].attack_range == 2
    np.testing.assert_array_equal(agents['agent5'].initial_position, np.array([4, 0]))

    actor = AttackActor(agents=agents)
    for agent in agents.values():
        agent.position = agent.initial_position

    assert actor.process_action(agents['agent0'], {'attack': True}).id == 'agent1'
    assert actor.process_action(agents['agent1'], {'attack': True}).id == 'agent0'
    assert actor.process_action(agents['agent2'], {'attack': True}).id == 'agent3'
    assert actor.process_action(agents['agent3'], {'attack': True}).id == 'agent2'
    assert actor.process_action(agents['agent4'], {'attack': True}) is None
    assert actor.process_action(agents['agent5'], {'attack': True}).id == 'agent2'

    agents['agent0'].is_alive = False
    agents['agent2'].is_alive = False

    assert actor.process_action(agents['agent1'], {'attack': True}) is None
    assert actor.process_action(agents['agent3'], {'attack': True}).id == 'agent4'
    assert actor.process_action(agents['agent4'], {'attack': True}) is None
    assert actor.process_action(agents['agent5'], {'attack': True}).id == 'agent4'


def test_position_team_based_attack_actor():
    agents = {
        'agent0': AttackingAgent(
            id='agent0', attack_range=1, initial_position=np.array([1, 1]), attack_strength=0.6,
            team=1
        ),
        'agent1': AttackingAgent(
            id='agent1', attack_range=1, initial_position=np.array([0, 1]), attack_strength=0.6,
            team=2
        ),
        'agent2': AttackingAgent(
            id='agent2', attack_range=1, initial_position=np.array([4, 2]), attack_strength=0.6,
            team=1
        ),
        'agent3': AttackingAgent(
            id='agent3', attack_range=1, initial_position=np.array([4, 3]), attack_strength=0.6,
            team=1
        ),
        'agent4': AttackingAgent(
            id='agent4', attack_range=0, initial_position=np.array([3, 2]), attack_strength=0.6,
            team=3
        ),
        'agent5': AttackingAgent(
            id='agent5', attack_range=2, initial_position=np.array([4, 0]), attack_strength=0.6,
            team=1
        ),
    }

    assert agents['agent0'].attack_range == 1
    assert agents['agent0'].team == 1
    np.testing.assert_array_equal(agents['agent0'].initial_position, np.array([1, 1]))
    assert agents['agent1'].attack_range == 1
    assert agents['agent1'].team == 2
    np.testing.assert_array_equal(agents['agent1'].initial_position, np.array([0, 1]))
    assert agents['agent2'].attack_range == 1
    assert agents['agent2'].team == 1
    np.testing.assert_array_equal(agents['agent2'].initial_position, np.array([4, 2]))
    assert agents['agent3'].attack_range == 1
    assert agents['agent3'].team == 1
    np.testing.assert_array_equal(agents['agent3'].initial_position, np.array([4, 3]))
    assert agents['agent4'].attack_range == 0
    assert agents['agent4'].team == 3
    np.testing.assert_array_equal(agents['agent4'].initial_position, np.array([3, 2]))
    assert agents['agent5'].attack_range == 2
    assert agents['agent5'].team == 1
    np.testing.assert_array_equal(agents['agent5'].initial_position, np.array([4, 0]))

    for agent in agents.values():
        agent.position = agent.initial_position
    actor = AttackActor(agents=agents, number_of_teams=3)

    assert actor.process_action(agents['agent0'], {'attack': True}).id == 'agent1'
    assert actor.process_action(agents['agent1'], {'attack': True}).id == 'agent0'
    assert actor.process_action(agents['agent2'], {'attack': True}).id == 'agent4'
    assert actor.process_action(agents['agent3'], {'attack': True}).id == 'agent4'
    assert actor.process_action(agents['agent4'], {'attack': True}) is None
    assert actor.process_action(agents['agent5'], {'attack': True}).id == 'agent4'

    agents['agent4'].is_alive = False
    agents['agent0'].is_alive = False

    assert actor.process_action(agents['agent1'], {'attack': True}) is None
    assert actor.process_action(agents['agent2'], {'attack': True}) is None
    assert actor.process_action(agents['agent3'], {'attack': True}) is None
    assert actor.process_action(agents['agent5'], {'attack': True}) is None


def test_attack_accuracy():
    np.random.seed(24)
    agents = {
        'agent0': AttackingAgent(
            id='agent0', attack_range=1, attack_strength=0, attack_accuracy=0,
            initial_position=np.array([1,1])
        ),
        'agent1': AttackingAgent(
            id='agent1', attack_range=0, attack_strength=0, initial_position=np.array([0,0])
        )
    }

    assert agents['agent0'].attack_accuracy == 0
    assert agents['agent1'].attack_accuracy == 1

    for agent in agents.values():
        agent.position = agent.initial_position

    actor = AttackActor(agents=agents)
    # Action failed because low accuracy
    assert actor.process_action(agents['agent0'], {'attack': True}) is None

    agents['agent0'].attack_accuracy = 0.5
    assert actor.process_action(agents['agent0'], {'attack': True}) is None
    assert actor.process_action(agents['agent0'], {'attack': True}) is None
    assert actor.process_action(agents['agent0'], {'attack': True}).id == 'agent1'
    assert actor.process_action(agents['agent0'], {'attack': True}).id == 'agent1'
    assert actor.process_action(agents['agent0'], {'attack': True}) is None
    assert actor.process_action(agents['agent0'], {'attack': True}) is None


def test_team_matrix():
    agents = {
        'agent0': AttackingAgent(
            id='agent0', attack_range=1, initial_position=np.array([1, 1]), attack_strength=0.6,
            team=1
        ),
        'agent1': AttackingAgent(
            id='agent1', attack_range=4, initial_position=np.array([0, 1]), attack_strength=0.6,
            team=2
        ),
        'agent2': AttackingAgent(
            id='agent2', attack_range=1, initial_position=np.array([4, 2]), attack_strength=0.6,
            team=1
        ),
        'agent3': AttackingAgent(
            id='agent3', attack_range=1, initial_position=np.array([4, 3]), attack_strength=0.6
        ),
        'agent4': AttackingAgent(
            id='agent4', attack_range=1, initial_position=np.array([3, 2]), attack_strength=0.6,
            team=3
        ),
        'agent5': AttackingAgent(
            id='agent5', attack_range=1, initial_position=np.array([4, 0]), attack_strength=0.6,
            team=1
        ),
    }

    for agent in agents.values():
        agent.position = agent.initial_position

    team_attack_matrix = np.zeros((4,4))
    team_attack_matrix[0, :] = 1
    team_attack_matrix[1, 0] = 1
    team_attack_matrix[2, 3] = 1

    actor = AttackActor(agents=agents, number_of_teams=3, team_attack_matrix=team_attack_matrix)
    assert actor.process_action(agents['agent0'], {'attack': True}) is None
    assert actor.process_action(agents['agent1'], {'attack': True}).id == 'agent4'
    assert actor.process_action(agents['agent2'], {'attack': True}).id == 'agent3'
    assert actor.process_action(agents['agent3'], {'attack': True}).id == 'agent2'
    assert actor.process_action(agents['agent4'], {'attack': True}) is None
    assert actor.process_action(agents['agent5'], {'attack': True}) is None



import numpy as np

from abmarl.sim.components.state import GridPositionState, LifeState
from abmarl.sim.components.observer import GridPositionBasedObserver, \
    GridPositionTeamBasedObserver, RelativePositionObserver
from abmarl.sim.components.agent import PositionObservingAgent, AgentObservingAgent, \
    ComponentAgent


class PositionTestAgent(PositionObservingAgent, AgentObservingAgent): pass
class PositionTeamTestAgent(PositionObservingAgent, AgentObservingAgent): pass
class PositionTeamNoViewTestAgent(ComponentAgent): pass


def test_grid_position_observer():
    agents = {
        'agent0': PositionTestAgent(id='agent0', initial_position=np.array([0, 0]), agent_view=1),
        'agent1': PositionTestAgent(id='agent1', initial_position=np.array([2, 2]), agent_view=2),
        'agent2': PositionTestAgent(id='agent2', initial_position=np.array([3, 2]), agent_view=3),
        'agent3': PositionTestAgent(id='agent3', initial_position=np.array([1, 4]), agent_view=4),
        'agent4': ComponentAgent(id='agent4', initial_position=np.array([1, 4])),
    }

    state = GridPositionState(agents=agents, region=5)
    life = LifeState(agents=agents)
    observer = GridPositionBasedObserver(position_state=state, agents=agents)
    state.reset()
    life.reset()

    np.testing.assert_array_equal(observer.get_obs(agents['agent0'])['position'], np.array([
        [-1., -1., -1.],
        [-1.,  0.,  0.],
        [-1.,  0.,  0.],
    ]))
    np.testing.assert_array_equal(observer.get_obs(agents['agent1'])['position'], np.array([
        [1., 0., 0., 0., 0.],
        [0., 0., 0., 0., 1.],
        [0., 0., 0., 0., 0.],
        [0., 0., 1., 0., 0.],
        [0., 0., 0., 0., 0.],
    ]))
    np.testing.assert_array_equal(observer.get_obs(agents['agent2'])['position'], np.array([
        [-1.,  1.,  0.,  0.,  0.,  0., -1.],
        [-1.,  0.,  0.,  0.,  0.,  1., -1.],
        [-1.,  0.,  0.,  1.,  0.,  0., -1.],
        [-1.,  0.,  0.,  0.,  0.,  0., -1.],
        [-1.,  0.,  0.,  0.,  0.,  0., -1.],
        [-1., -1., -1., -1., -1., -1., -1.],
        [-1., -1., -1., -1., -1., -1., -1.],
    ]))
    np.testing.assert_array_equal(observer.get_obs(agents['agent3'])['position'], np.array([
        [-1., -1., -1., -1., -1., -1., -1., -1., -1.],
        [-1., -1., -1., -1., -1., -1., -1., -1., -1.],
        [-1., -1., -1., -1., -1., -1., -1., -1., -1.],
        [ 1.,  0.,  0.,  0.,  0., -1., -1., -1., -1.],
        [ 0.,  0.,  0.,  0.,  1., -1., -1., -1., -1.],
        [ 0.,  0.,  1.,  0.,  0., -1., -1., -1., -1.],
        [ 0.,  0.,  1.,  0.,  0., -1., -1., -1., -1.],
        [ 0.,  0.,  0.,  0.,  0., -1., -1., -1., -1.],
        [-1., -1., -1., -1., -1., -1., -1., -1., -1.],
    ]))
    assert observer.get_obs(agents['agent4']) == {}


def test_grid_team_position_observer():
    agents = {
        'agent0': PositionTeamTestAgent(
            id='agent0', team=1, initial_position=np.array([0, 0]), agent_view=1
        ),
        'agent1': PositionTeamNoViewTestAgent(
            id='agent1', team=1, initial_position=np.array([0, 0])
        ),
        'agent2': PositionTeamTestAgent(
            id='agent2', team=1, initial_position=np.array([2, 2]), agent_view=2
        ),
        'agent3': PositionTeamTestAgent(
            id='agent3', team=2, initial_position=np.array([3, 2]), agent_view=3
        ),
        'agent4': PositionTeamTestAgent(
            id='agent4', team=2, initial_position=np.array([1, 4]), agent_view=4
        ),
        'agent5': PositionTeamNoViewTestAgent(
            id='agent5', team=2, initial_position=np.array([1, 4])
        ),
        'agent6': PositionTeamNoViewTestAgent(
            id='agent6', team=2, initial_position=np.array([1, 4])
        ),
        'agent7': PositionTeamTestAgent(
            id='agent7', team=3, initial_position=np.array([1, 4]), agent_view=2
        ),
    }
    for agent in agents.values():
        agent.position = agent.initial_position

    state = GridPositionState(agents=agents, region=5)
    life = LifeState(agents=agents)
    observer = GridPositionTeamBasedObserver(
        position_state=state, number_of_teams=3, agents=agents
    )
    state.reset()
    life.reset()

    np.testing.assert_array_equal(observer.get_obs(agents['agent0'])['position'][:,:,1], np.array([
        [-1., -1., -1.],
        [-1.,  1.,  0.],
        [-1.,  0.,  0.],
    ]))
    np.testing.assert_array_equal(observer.get_obs(agents['agent0'])['position'][:,:,2], np.array([
        [-1., -1., -1.],
        [-1.,  0.,  0.],
        [-1.,  0.,  0.],
    ]))
    np.testing.assert_array_equal(observer.get_obs(agents['agent0'])['position'][:,:,3], np.array([
        [-1., -1., -1.],
        [-1.,  0.,  0.],
        [-1.,  0.,  0.],
    ]))

    np.testing.assert_array_equal(observer.get_obs(agents['agent2'])['position'][:,:,1], np.array([
        [2., 0., 0., 0., 0.],
        [0., 0., 0., 0., 0.],
        [0., 0., 0., 0., 0.],
        [0., 0., 0., 0., 0.],
        [0., 0., 0., 0., 0.],
    ]))
    np.testing.assert_array_equal(observer.get_obs(agents['agent2'])['position'][:,:,2], np.array([
        [0., 0., 0., 0., 0.],
        [0., 0., 0., 0., 3.],
        [0., 0., 0., 0., 0.],
        [0., 0., 1., 0., 0.],
        [0., 0., 0., 0., 0.],
    ]))
    np.testing.assert_array_equal(observer.get_obs(agents['agent2'])['position'][:,:,3], np.array([
        [0., 0., 0., 0., 0.],
        [0., 0., 0., 0., 1.],
        [0., 0., 0., 0., 0.],
        [0., 0., 0., 0., 0.],
        [0., 0., 0., 0., 0.],
    ]))

    np.testing.assert_array_equal(observer.get_obs(agents['agent3'])['position'][:,:,1], np.array([
        [-1.,  2.,  0.,  0.,  0.,  0., -1.],
        [-1.,  0.,  0.,  0.,  0.,  0., -1.],
        [-1.,  0.,  0.,  1.,  0.,  0., -1.],
        [-1.,  0.,  0.,  0.,  0.,  0., -1.],
        [-1.,  0.,  0.,  0.,  0.,  0., -1.],
        [-1., -1., -1., -1., -1., -1., -1.],
        [-1., -1., -1., -1., -1., -1., -1.],
    ]))
    np.testing.assert_array_equal(observer.get_obs(agents['agent3'])['position'][:,:,2], np.array([
        [-1.,  0.,  0.,  0.,  0.,  0., -1.],
        [-1.,  0.,  0.,  0.,  0.,  3., -1.],
        [-1.,  0.,  0.,  0.,  0.,  0., -1.],
        [-1.,  0.,  0.,  0.,  0.,  0., -1.],
        [-1.,  0.,  0.,  0.,  0.,  0., -1.],
        [-1., -1., -1., -1., -1., -1., -1.],
        [-1., -1., -1., -1., -1., -1., -1.],
    ]))
    np.testing.assert_array_equal(observer.get_obs(agents['agent3'])['position'][:,:,3], np.array([
        [-1.,  0.,  0.,  0.,  0.,  0., -1.],
        [-1.,  0.,  0.,  0.,  0.,  1., -1.],
        [-1.,  0.,  0.,  0.,  0.,  0., -1.],
        [-1.,  0.,  0.,  0.,  0.,  0., -1.],
        [-1.,  0.,  0.,  0.,  0.,  0., -1.],
        [-1., -1., -1., -1., -1., -1., -1.],
        [-1., -1., -1., -1., -1., -1., -1.],
    ]))

    np.testing.assert_array_equal(observer.get_obs(agents['agent4'])['position'][:,:,1], np.array([
        [-1., -1., -1., -1., -1., -1., -1., -1., -1.],
        [-1., -1., -1., -1., -1., -1., -1., -1., -1.],
        [-1., -1., -1., -1., -1., -1., -1., -1., -1.],
        [ 2.,  0.,  0.,  0.,  0., -1., -1., -1., -1.],
        [ 0.,  0.,  0.,  0.,  0., -1., -1., -1., -1.],
        [ 0.,  0.,  1.,  0.,  0., -1., -1., -1., -1.],
        [ 0.,  0.,  0.,  0.,  0., -1., -1., -1., -1.],
        [ 0.,  0.,  0.,  0.,  0., -1., -1., -1., -1.],
        [-1., -1., -1., -1., -1., -1., -1., -1., -1.],
    ]))
    np.testing.assert_array_equal(observer.get_obs(agents['agent4'])['position'][:,:,2], np.array([
        [-1., -1., -1., -1., -1., -1., -1., -1., -1.],
        [-1., -1., -1., -1., -1., -1., -1., -1., -1.],
        [-1., -1., -1., -1., -1., -1., -1., -1., -1.],
        [ 0.,  0.,  0.,  0.,  0., -1., -1., -1., -1.],
        [ 0.,  0.,  0.,  0.,  2., -1., -1., -1., -1.],
        [ 0.,  0.,  0.,  0.,  0., -1., -1., -1., -1.],
        [ 0.,  0.,  1.,  0.,  0., -1., -1., -1., -1.],
        [ 0.,  0.,  0.,  0.,  0., -1., -1., -1., -1.],
        [-1., -1., -1., -1., -1., -1., -1., -1., -1.],
    ]))
    np.testing.assert_array_equal(observer.get_obs(agents['agent4'])['position'][:,:,3], np.array([
        [-1., -1., -1., -1., -1., -1., -1., -1., -1.],
        [-1., -1., -1., -1., -1., -1., -1., -1., -1.],
        [-1., -1., -1., -1., -1., -1., -1., -1., -1.],
        [ 0.,  0.,  0.,  0.,  0., -1., -1., -1., -1.],
        [ 0.,  0.,  0.,  0.,  1., -1., -1., -1., -1.],
        [ 0.,  0.,  0.,  0.,  0., -1., -1., -1., -1.],
        [ 0.,  0.,  0.,  0.,  0., -1., -1., -1., -1.],
        [ 0.,  0.,  0.,  0.,  0., -1., -1., -1., -1.],
        [-1., -1., -1., -1., -1., -1., -1., -1., -1.],
    ]))

    np.testing.assert_array_equal(observer.get_obs(agents['agent7'])['position'][:,:,1], np.array([
        [-1., -1., -1., -1., -1.],
        [ 0.,  0.,  0., -1., -1.],
        [ 0.,  0.,  0., -1., -1.],
        [ 1.,  0.,  0., -1., -1.],
        [ 0.,  0.,  0., -1., -1.],
    ]))
    np.testing.assert_array_equal(observer.get_obs(agents['agent7'])['position'][:,:,2], np.array([
        [-1., -1., -1., -1., -1.],
        [ 0.,  0.,  0., -1., -1.],
        [ 0.,  0.,  3., -1., -1.],
        [ 0.,  0.,  0., -1., -1.],
        [ 1.,  0.,  0., -1., -1.],
    ]))
    np.testing.assert_array_equal(observer.get_obs(agents['agent7'])['position'][:,:,3], np.array([
        [-1., -1., -1., -1., -1.],
        [ 0.,  0.,  0., -1., -1.],
        [ 0.,  0.,  0., -1., -1.],
        [ 0.,  0.,  0., -1., -1.],
        [ 0.,  0.,  0., -1., -1.],
    ]))


def test_relative_position_observer():
    agents = {
        'agent0': PositionTestAgent(id='agent0', initial_position=np.array([0, 0]), agent_view=1),
        'agent1': PositionTestAgent(id='agent1', initial_position=np.array([2, 2]), agent_view=2),
        'agent2': PositionTestAgent(id='agent2', initial_position=np.array([3, 2]), agent_view=3),
        'agent3': PositionTestAgent(id='agent3', initial_position=np.array([1, 4]), agent_view=4),
        'agent4': ComponentAgent(id='agent4', initial_position=np.array([1, 4])),
    }

    state = GridPositionState(agents=agents, region=5)
    observer = RelativePositionObserver(position_state=state, agents=agents)
    state.reset()

    assert observer.get_obs(agents['agent0'])['relative_position']['agent1'][0] == 2
    assert observer.get_obs(agents['agent0'])['relative_position']['agent1'][1] == 2
    assert observer.get_obs(agents['agent0'])['relative_position']['agent2'][0] == 3
    assert observer.get_obs(agents['agent0'])['relative_position']['agent2'][1] == 2
    assert observer.get_obs(agents['agent0'])['relative_position']['agent3'][0] == 1
    assert observer.get_obs(agents['agent0'])['relative_position']['agent3'][1] == 4
    assert observer.get_obs(agents['agent0'])['relative_position']['agent4'][0] == 1
    assert observer.get_obs(agents['agent0'])['relative_position']['agent4'][1] == 4

    assert observer.get_obs(agents['agent1'])['relative_position']['agent0'][0] == -2
    assert observer.get_obs(agents['agent1'])['relative_position']['agent0'][1] == -2
    assert observer.get_obs(agents['agent1'])['relative_position']['agent2'][0] == 1
    assert observer.get_obs(agents['agent1'])['relative_position']['agent2'][1] == 0
    assert observer.get_obs(agents['agent1'])['relative_position']['agent3'][0] == -1
    assert observer.get_obs(agents['agent1'])['relative_position']['agent3'][1] == 2
    assert observer.get_obs(agents['agent1'])['relative_position']['agent4'][0] == -1
    assert observer.get_obs(agents['agent1'])['relative_position']['agent4'][1] == 2

    assert observer.get_obs(agents['agent2'])['relative_position']['agent0'][0] == -3
    assert observer.get_obs(agents['agent2'])['relative_position']['agent0'][1] == -2
    assert observer.get_obs(agents['agent2'])['relative_position']['agent1'][0] == -1
    assert observer.get_obs(agents['agent2'])['relative_position']['agent1'][1] == 0
    assert observer.get_obs(agents['agent2'])['relative_position']['agent3'][0] == -2
    assert observer.get_obs(agents['agent2'])['relative_position']['agent3'][1] == 2
    assert observer.get_obs(agents['agent2'])['relative_position']['agent4'][0] == -2
    assert observer.get_obs(agents['agent2'])['relative_position']['agent4'][1] == 2

    assert observer.get_obs(agents['agent3'])['relative_position']['agent0'][0] == -1
    assert observer.get_obs(agents['agent3'])['relative_position']['agent0'][1] == -4
    assert observer.get_obs(agents['agent3'])['relative_position']['agent1'][0] == 1
    assert observer.get_obs(agents['agent3'])['relative_position']['agent1'][1] == -2
    assert observer.get_obs(agents['agent3'])['relative_position']['agent2'][0] == 2
    assert observer.get_obs(agents['agent3'])['relative_position']['agent2'][1] == -2
    assert observer.get_obs(agents['agent3'])['relative_position']['agent4'][0] == 0
    assert observer.get_obs(agents['agent3'])['relative_position']['agent4'][1] == 0



import numpy as np

from abmarl.sim.components.state import VelocityState, ContinuousPositionState
from abmarl.sim.components.actor import AccelerationMovementActor, ContinuousCollisionActor
from abmarl.sim.components.actor import VelocityAgent, AcceleratingAgent, CollisionAgent


class ParticleAgent(VelocityAgent, AcceleratingAgent, CollisionAgent): pass


def test_collision():

    agents = {
        'agent0': ParticleAgent(
            id='agent0', max_acceleration=0, max_speed=10, size=1, mass=1,
            initial_velocity=np.array([1, 1]), initial_position=np.array([1,1])
        ),
        'agent1': ParticleAgent(
            id='agent1', max_acceleration=0, max_speed=10, size=1, mass=1,
            initial_velocity=np.array([-1, 1]), initial_position=np.array([4, 1])
        )
    }

    position_state = ContinuousPositionState(region=10, agents=agents)
    velocity_state = VelocityState(agents=agents, friction=0.0)
    position_state.reset()
    velocity_state.reset()

    movement_actor = AccelerationMovementActor(
        position_state=position_state, velocity_state=velocity_state, agents=agents
    )
    collision_actor = ContinuousCollisionActor(
        position_state=position_state, velocity_state=velocity_state, agents=agents
    )

    np.testing.assert_array_equal(
        movement_actor.process_action(agents['agent0'], {'accelerate': np.zeros(2)}),
        np.array([1., 1.])
    )
    np.testing.assert_array_equal(agents['agent0'].position, np.array([2., 2.]))
    np.testing.assert_array_equal(agents['agent0'].velocity, np.array([1., 1.]))
    np.testing.assert_array_equal(
        movement_actor.process_action(agents['agent1'], {'accelerate': np.zeros(2)}),
        np.array([-1., 1.])
    )
    np.testing.assert_array_equal(agents['agent1'].position, np.array([3., 2.]))
    np.testing.assert_array_equal(agents['agent1'].velocity, np.array([-1., 1.]))

    collision_actor.detect_collisions_and_modify_states()
    np.testing.assert_array_equal(agents['agent0'].position, np.array([1.5, 1.5]))
    np.testing.assert_array_equal(agents['agent1'].position, np.array([3.5, 1.5]))
    np.testing.assert_array_equal(agents['agent0'].velocity, np.array([-1., 1.]))
    np.testing.assert_array_equal(agents['agent1'].velocity, np.array([1., 1.]))

    np.testing.assert_array_equal(
        movement_actor.process_action(agents['agent0'], {'accelerate': np.zeros(2)}),
        np.array([-1., 1.])
    )
    np.testing.assert_array_equal(
        movement_actor.process_action(agents['agent1'], {'accelerate': np.zeros(2)}),
        np.array([1., 1.])
    )



from abc import ABC, abstractmethod

import numpy as np

from abmarl.sim.components.agent import SpeedAngleAgent, VelocityAgent, CollisionAgent, \
    BroadcastingAgent


# --------------------- #
# --- Communication --- #
# --------------------- #

class BroadcastState:
    """
    Tracks which agents have broadcasted in this step.

    agents (dict):
        The dictionary of agents.
    """
    def __init__(self, agents=None, **kwargs):
        self.agents = agents

    def reset(self, **kwargs):
        """
        Reset the broadcasting state of all applicable agents.
        """
        for agent in self.agents.values():
            if isinstance(agent, BroadcastingAgent):
                agent.broadcasting = False

    def set_broadcast(self, agent, _broadcast):
        """
        Set the broadcasting state of the agent.
        """
        if isinstance(agent, BroadcastingAgent):
            agent.broadcasting = _broadcast

    def modify_broadcast(self, agent, value):
        """
        Set the broadcasting state of the agent.
        """
        self.set_broadcast(agent, value)


# ----------------------- #
# --- Health and Life --- #
# ----------------------- #

class LifeState:
    """
    Agents can die if their health falls below their minimal health value. Health
    can decrease in a number of interactions. This simulation provides an entropy
    that indicates how much health an agent loses when apply_entropy is called.
    This is a generic entropy for the step. If you want to specify health changes
    for specific actions, such as being attacked or harvesting, you must write
    it in the simulation.

    agents (dict):
        Dictionary of agents.

    entropy (float):
        The amount of health that is depleted from an agent whenever apply_entropy
        is called.
    """
    def __init__(self, agents=None, entropy=0.1, **kwargs):
        assert type(agents) is dict, "Agents must be a dict"
        self.agents = agents
        self.entropy = entropy

    def reset(self, **kwargs):
        """
        Reset the health and life state of all applicable agents.
        """
        for agent in self.agents.values():
            if agent.initial_health is not None:
                agent.health = agent.initial_health
            else:
                agent.health = np.random.uniform(agent.min_health, agent.max_health)
            agent.is_alive = True

    def set_health(self, agent, _health):
        """
        Set the health of an agent to a specific value, bounded by the agent's
        min and max health-value. If that value is less than the agent's health,
        then the agent dies.
        """
        if _health <= agent.min_health:
            agent.health = 0
            agent.is_alive = False
        elif _health >= agent.max_health:
            agent.health = agent.max_health
        else:
            agent.health = _health

    def modify_health(self, agent, value):
        """
        Add some value to the health of the agent.
        """
        self.set_health(agent, agent.health + value)

    def apply_entropy(self, agent, **kwargs):
        """
        Apply entropy to the agent, decreasing its health by a small amount.
        """
        self.modify_health(agent, -self.entropy, **kwargs)


# ----------------------------- #
# --- Position and Movement --- #
# ----------------------------- #

class PositionState(ABC):
    """
    Manages the agents' positions.

    region (int):
        The size of the simulation map.

    agents (dict):
        The dictionary of agents.
    """
    def __init__(self, region=None, agents=None, **kwargs):
        assert type(region) is int, "Region must be an integer."
        self.region = region
        assert type(agents) is dict, "agents must be a dict"
        self.agents = agents

    def reset(self, **kwargs):
        """
        Reset the agents' positions. If the agents were created with a starting
        position, then use that. Otherwise, randomly assign a position in the region.
        """
        # Invalidate all the agents' positions from last episode
        for agent in self.agents.values():
            agent.position = None

        for agent in self.agents.values():
            if agent.initial_position is not None:
                agent.position = agent.initial_position
            else:
                self.random_reset(agent)

    @abstractmethod
    def random_reset(self, agent, **kwargs):
        """
        Reset the agents' positions. Child classes implement this according to their
        specs. For example, GridPositionState assigns random integers as the position,
        whereas ContinuousPositionState assigns random numbers.
        """
        pass

    @abstractmethod
    def set_position(self, agent, position, **kwargs):
        """
        Set the position of the agents. Child classes implement.
        """
        pass

    def modify_position(self, agent, value, **kwargs):
        """
        Add some value to the position of the agent.
        """
        self.set_position(agent, agent.position + value)


class GridPositionState(PositionState):
    """
    Agents are positioned in a grid and cannot go outside of the region. Positions
    are a 2-element numpy array, where the first element is the grid-row from top
    to bottom and the second is the grid-column from left to right.
    """
    def set_position(self, agent, _position, **kwargs):
        """
        Set the agent's position to the incoming value only if the new position
        is within the region.
        """
        if 0 <= _position[0] < self.region and 0 <= _position[1] < self.region:
            agent.position = _position

    def random_reset(self, agent, **kwargs):
        """
        Set the agents' random positions as integers within the region.
        """
        agent.position = np.random.randint(0, self.region, 2)


class ContinuousPositionState(PositionState):
    """
    Agents are positioned in a continuous space and can go outside the bounds
    of the region. Positions are a 2-element array, where the first element is
    the x-location and the second is the y-location.
    """
    def __init__(self, reset_attempts=100, **kwargs):
        super().__init__(**kwargs)
        self.reset_attempts = reset_attempts

    def set_position(self, agent, _position, **kwargs):
        """
        Set the agent's position to the incoming value.
        """
        agent.position = _position

    def random_reset(self, agent, **kwargs):
        """
        Set the agents' random positions as numbers within the region.
        """
        if isinstance(agent, CollisionAgent):
            for _ in range(self.reset_attempts):
                potential_position = np.random.uniform(0, self.region, 2)
                collision = False
                for other in self.agents.values():
                    if other.id != agent.id and \
                       isinstance(other, CollisionAgent) and \
                       other.position is not None and \
                       np.linalg.norm(other.position - potential_position) < \
                                     (other.size + agent.size):
                        collision = True
                        break
                if not collision:
                    agent.position = potential_position
                    return
            raise Exception("Could not fit all the agents in the region without collisions")
        else:
            agent.position = np.random.uniform(0, self.region, 2)


class SpeedAngleState:
    """
    Manages the agents' speed, banking angles, and ground angles.
    """
    def __init__(self, agents=None, **kwargs):
        self.agents = agents

    def reset(self, **kwargs):
        """
        Reset the agents' speeds and ground angles.
        """
        for agent in self.agents.values():
            if isinstance(agent, SpeedAngleAgent):
                # Reset agent speed
                if agent.initial_speed is not None:
                    agent.speed = agent.initial_speed
                else:
                    agent.speed = np.random.uniform(agent.min_speed, agent.max_speed)

                # Reset agent banking angle
                if agent.initial_banking_angle is not None:
                    agent.banking_angle = agent.initial_banking_angle
                else:
                    agent.banking_angle = np.random.uniform(
                        -agent.max_banking_angle, agent.max_banking_angle
                    )

                # Reset agent ground angle
                if agent.initial_ground_angle is not None:
                    agent.ground_angle = agent.initial_ground_angle
                else:
                    agent.ground_angle = np.random.uniform(0, 360)

    def set_speed(self, agent, _speed, **kwargs):
        """
        Set the agent's speed if it is between its min and max speed.
        """
        if isinstance(agent, SpeedAngleAgent):
            if agent.min_speed <= _speed <= agent.max_speed:
                agent.speed = _speed

    def modify_speed(self, agent, value, **kwargs):
        """
        Modify the agent's speed.
        """
        if isinstance(agent, SpeedAngleAgent):
            self.set_speed(agent, agent.speed + value)

    def set_banking_angle(self, agent, _banking_angle, **kwargs):
        """
        Set the agent's banking angle if it is between its min and max angle.
        """
        if isinstance(agent, SpeedAngleAgent):
            if abs(_banking_angle) <= agent.max_banking_angle:
                agent.banking_angle = _banking_angle
                self.modify_ground_angle(agent, agent.banking_angle)

    def modify_banking_angle(self, agent, value, **kwargs):
        """
        Modify the agent's banking angle.
        """
        if isinstance(agent, SpeedAngleAgent):
            self.set_banking_angle(agent, agent.banking_angle + value)

    def set_ground_angle(self, agent, _ground_angle, **kwargs):
        """
        Set the agent's ground angle, which will be modded to fall between 0 and
        360.
        """
        if isinstance(agent, SpeedAngleAgent):
            agent.ground_angle = _ground_angle % 360

    def modify_ground_angle(self, agent, value, **kwargs):
        """
        Modify the agent's ground angle.
        """
        if isinstance(agent, SpeedAngleAgent):
            self.set_ground_angle(agent, agent.ground_angle + value)


class VelocityState:
    """
    Manages the agents' velocities.
    """
    def __init__(self, agents=None, friction=0.05, **kwargs):
        self.agents = agents
        self.friction = friction

    def reset(self, **kwargs):
        """
        Reset the agents' velocities.
        """
        for agent in self.agents.values():
            if isinstance(agent, VelocityAgent):
                # Reset the agent's velocity
                if agent.initial_velocity is not None:
                    agent.velocity = agent.initial_velocity
                else:
                    agent.velocity = np.random.uniform(-agent.max_speed, agent.max_speed, (2,))

    def set_velocity(self, agent, _velocity, **kwargs):
        """
        Set the agent's velocity if it is within its max speed.
        """
        if isinstance(agent, VelocityAgent):
            vel_norm = np.linalg.norm(_velocity)
            if vel_norm < agent.max_speed:
                agent.velocity = _velocity
            else:
                agent.velocity = _velocity / vel_norm * agent.max_speed

    def modify_velocity(self, agent, value, **kwargs):
        """
        Modify the agent's velocity.
        """
        if isinstance(agent, VelocityAgent):
            self.set_velocity(agent, agent.velocity + value, **kwargs)

    def apply_friction(self, agent, **kwargs):
        """
        Apply friction to the agent's movement, decreasing its speed by a small amount.
        """
        if isinstance(agent, VelocityAgent):
            old_speed = np.linalg.norm(agent.velocity)
            new_speed = old_speed - self.friction
            if new_speed <= 0:
                agent.velocity = np.zeros(2)
            else:
                agent.velocity *= new_speed / old_speed


# -------------------------------- #
# --- Resources and Harvesting --- #
# -------------------------------- #

class GridResourceState:
    """
    Resources exist in the cells of the grid. The grid is populated with resources
    between the min and max value on some coverage of the region at reset time.
    If original resources is specified, then reset will set the resources back
    to that original value. This component supports resource depletion: if a resource falls below
    the minimum value, it will not regrow. Agents can harvest resources from the cell they occupy.
    Agents can observe the resources in a grid-like observation surrounding their positions.

    An agent can harvest up to its max harvest value on the cell it occupies. It
    can observe the resources in a grid surrounding its position, up to its view
    distance.

    agents (dict):
        The dictionary of agents.

    region (int):
        The size of the region

    coverage (float):
        The ratio of the region that should start with resources.

    min_value (float):
        The minimum value a resource can have before it cannot grow back. This is
        different from the absolute minimum value, 0, which indicates that there
        are no resources in the cell.

    max_value (float):
        The maximum value a resource can have.

    regrow_rate (float):
        The rate at which resources regrow.

    initial_resources (np.array):
        Instead of specifying the above resource-related parameters, we can provide
        an initial state of the resources. At reset time, the resources will be
        set to these original resources. Otherwise, the resources will be set
        to random values between the min and max value up to some coverage of the
        region.
    """
    def __init__(self, agents=None, region=None, coverage=0.75, min_value=0.1, max_value=1.0,
                 regrow_rate=0.04, initial_resources=None, **kwargs):
        self.initial_resources = initial_resources
        if self.initial_resources is None:
            assert type(region) is int, "Region must be an integer."
            self.region = region
        else:
            self.region = self.initial_resources.shape[0]
        self.min_value = min_value
        self.max_value = max_value
        self.regrow_rate = regrow_rate
        self.coverage = coverage

        assert type(agents) is dict, "agents must be a dict"
        self.agents = agents

    def reset(self, **kwargs):
        """
        Reset the resources. If original resources is specified, then the resources
        will be reset back to this original value. Otherwise, the resources will
        be randomly generated values between the min and max value up to some coverage
        of the region.
        """
        if self.initial_resources is not None:
            self.resources = self.initial_resources
        else:
            coverage_filter = np.zeros((self.region, self.region))
            coverage_filter[
                np.random.uniform(0, 1, (self.region, self.region)) < self.coverage
            ] = 1.
            self.resources = np.multiply(
                np.random.uniform(self.min_value, self.max_value, (self.region, self.region)),
                coverage_filter
            )

    def set_resources(self, location, value, **kwargs):
        """
        Set the resource at a certain location to a value, bounded between 0 and
        the maximum resource value.
        """
        assert type(location) is tuple
        if value <= 0:
            self.resources[location] = 0
        elif value >= self.max_value:
            self.resources[location] = self.max_value
        else:
            self.resources[location] = value

    def modify_resources(self, location, value, **kwargs):
        """
        Add some value to the resource at a certain location.
        """
        assert type(location) is tuple
        self.set_resources(location, self.resources[location] + value, **kwargs)

    def regrow(self, **kwargs):
        """
        Regrow the resources according to the regrow_rate.
        """
        self.resources[self.resources >= self.min_value] += self.regrow_rate
        self.resources[self.resources >= self.max_value] = self.max_value
